from __future__ import annotations

from osaft.core.backgroundfields import BackgroundField, WaveType
from osaft.core.basecomposite import BaseSphereFrequencyComposite
from osaft.core.fluids import InviscidFluid
from osaft.core.frequency import Frequency
from osaft.core.geometries import Sphere
from osaft.core.variable import ActiveVariable
from osaft.solutions.base_solution import BaseSolution


class BaseYosioka(BaseSphereFrequencyComposite, BaseSolution):
    """Base class for Yosioka & Kawasima (1955)

    :param f: Frequency [Hz]
    :param R_0: Radius of the particle [m]
    :param rho_s: Density of the fluid-like sphere [kg/m^3]
    :param c_s: Speed of sound of the particle [m/s]
    :param rho_f: Density of the fluid [kg/m^3]
    :param c_f: Speed of sound in the fluid [m/s]
    :param p_0: Pressure amplitude of the incident field [Pa]
    :param position: Position in the standing wave field [rad]
    :param wave_type: Either in or progressive wave [-]
    """

    # Supported wave_type for the class
    supported_wavetypes = [WaveType.STANDING, WaveType.TRAVELLING]

    def __init__(
        self,
        f: Frequency | float | int,
        R_0: Sphere | float | int,
        rho_s: float,
        c_s: float,
        rho_f: float,
        c_f: float,
        p_0: float,
        wave_type: WaveType,
        position: None | float = None,
    ) -> None:

        """Constructor method"""
        # init of parent class
        BaseSphereFrequencyComposite.__init__(self, f, R_0)
        BaseSolution.__init__(self, "Yosioka1955")

        # Initialize Components
        self.scatterer = InviscidFluid(self.frequency, rho_s, c_s)
        self.fluid = InviscidFluid(self.frequency, rho_f, c_f)
        self.field = BackgroundField(self.fluid, p_0, wave_type, position)

        # Dependent variables
        self._lambda_rho = ActiveVariable(
            self._compute_lambda_rho,
            "ratio of densities",
        )
        self._sigma = ActiveVariable(
            self._compute_sigma,
            "ratio of speed of sounds",
        )

        self._x_f = ActiveVariable(
            self._compute_x_f,
            "dimensionless fluid wavenumber",
        )
        self._x_s = ActiveVariable(
            self._compute_x_s,
            "dimensionless particle wavenumber",
        )

        # Dependencies
        self._lambda_rho.is_computed_by(
            self.scatterer._rho_f,
            self.fluid._rho_f,
        )
        self._sigma.is_computed_by(self.scatterer._c_f, self.fluid._c_f)
        self._x_f.is_computed_by(self.sphere._R_0, self.fluid._k_f)
        self._x_s.is_computed_by(self.sphere._R_0, self.scatterer._k_f)

    # -------------------------------------------------------------------------
    # Getters for Dependent Variables
    # -------------------------------------------------------------------------

    @property
    def sigma(self) -> float:
        """ratio of speeds of sound [-]."""
        return self._sigma.value

    @property
    def lambda_rho(self) -> float:
        """ratio of densities [-]"""
        return self._lambda_rho.value

    @property
    def x_f(self) -> float:
        """dimensionless wavenumber in the fluid [-]"""
        return self._x_f.value

    @property
    def x_s(self) -> float:
        """dimensionless wavenumber in the particle"""
        return self._x_s.value

    # -------------------------------------------------------------------------
    # Wrappers for Independent Field Attributes
    # -------------------------------------------------------------------------

    @property
    def position(self) -> float:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.position`
        """
        return self.field.position

    @position.setter
    def position(self, value: float) -> None:
        self.field.position = value

    @property
    def p_0(self) -> float:
        """Wraps to :attr:`osaft.core.backgroundfields.BackgroundField.p_0`"""
        return self.field.p_0

    @p_0.setter
    def p_0(self, value: float) -> None:
        self.field.p_0 = value

    @property
    def wave_type(self) -> WaveType:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.wave_type`
        """
        return self.field.wave_type

    @wave_type.setter
    def wave_type(self, value: WaveType) -> None:
        self.field.wave_type = value

    # -------------------------------------------------------------------------
    # Wrappers for Independent Scatterer Attributes
    # -------------------------------------------------------------------------

    @property
    def rho_s(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.InviscidFluid.rho_f`"""
        return self.scatterer.rho_f

    @rho_s.setter
    def rho_s(self, value: float) -> None:
        self.scatterer.rho_f = value

    @property
    def c_s(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.InviscidFluid.c_f`"""
        return self.scatterer.c_f

    @c_s.setter
    def c_s(self, value: float) -> None:
        self.scatterer.c_f = value

    # -------------------------------------------------------------------------
    # Wrappers for Independent Fluid Attributes
    # -------------------------------------------------------------------------

    @property
    def rho_f(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.InviscidFluid.rho_f`"""
        return self.fluid.rho_f

    @rho_f.setter
    def rho_f(self, value: float) -> None:
        self.fluid.rho_f = value

    @property
    def c_f(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.InviscidFluid.c_f`"""
        return self.fluid.c_f

    @c_f.setter
    def c_f(self, value: float) -> None:
        self.fluid.c_f = value

    # -------------------------------------------------------------------------
    # Wrappers for Dependent Fluid Attributes
    # -------------------------------------------------------------------------

    @property
    def kappa_f(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.InviscidFluid.kappa_f`"""
        return self.fluid.kappa_f

    @property
    def k_f(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.InviscidFluid.k_f`"""
        return self.fluid.k_f

    # -------------------------------------------------------------------------
    # Wrappers for Dependent Scatterer Attributes
    # -------------------------------------------------------------------------

    @property
    def k_s(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.InviscidFluid.k_f`"""
        return self.scatterer.k_f

    # -------------------------------------------------------------------------
    # Wrappers for Dependent Field Attributes
    # -------------------------------------------------------------------------

    @property
    def I_ac(self) -> float:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.I_ac`
        """
        return self.field.I_ac

    @property
    def E_ac(self) -> float:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.E_ac`
        """
        return self.field.E_ac

    def A_in(self, n) -> complex:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.A_in`
        """
        return self.field.A_in(n)

    # -------------------------------------------------------------------------
    # Dependent Variables Methods
    # -------------------------------------------------------------------------

    def _compute_sigma(self) -> float:
        """Compute sigma (ratio of speed of sound)"""
        return self.c_s / self.c_f

    def _compute_lambda_rho(self) -> float:
        """Compute lambda (ratio of densities)"""
        return self.rho_s / self.rho_f

    def _compute_x_f(self) -> float:
        """Compute x_f"""
        return self.R_0 * self.k_f

    def _compute_x_s(self) -> float:
        """Compute x_s"""
        return self.R_0 * self.k_s


if __name__ == "__main__":
    pass



import unittest

from osaft.core.functions import BesselFunctions as Bf
from osaft.tests.basetest_numeric import BaseTestSolutions
from osaft.tests.basetest_scattering import HelperScattering
from osaft.tests.solution_factory import SolutionFactory


class TestScattering(BaseTestSolutions, HelperScattering):
    def setUp(self) -> None:

        BaseTestSolutions.setUp(self)

        self._v_boundary_conditions = 1e-9

        self.cls = SolutionFactory().hasegawa_1969_scattering()

        self.list_cls = [self.cls]

    # -------------------------------------------------------------------------
    # Methods
    # -------------------------------------------------------------------------

    def c_n(self, n: int) -> complex:

        nu = self.cls.nu_s
        k_f = self.cls.k_f
        k_s_1 = self.cls.k_s_l
        k_s_2 = self.cls.k_s_t
        x_f = self.R_0 * k_f
        x_s_1 = self.R_0 * k_s_1
        x_s_2 = self.R_0 * k_s_2
        j = lambda x: Bf.besselj(n, x)
        dj = lambda x: Bf.d1_besselj(n, x)
        ddj = lambda x: Bf.d2_besselj(n, x)
        h = lambda x: Bf.hankelh2(n, x)
        dh = lambda x: Bf.d1_hankelh2(n, x)

        frac1 = (x_s_1 * dj(x_s_1)) / (x_s_1 * dj(x_s_1) - j(x_s_1))
        frac2 = (
            2
            * n
            * (n + 1)
            * j(x_s_2)
            / ((n + 2) * (n - 1) * j(x_s_2) + x_s_2**2 * ddj(x_s_2))
        )
        frac3 = (
            x_s_1**2
            * (nu / (1 - 2 * nu) * j(x_s_1) - ddj(x_s_1))
            / (x_s_1 * dj(x_s_1) - j(x_s_1))
        )
        frac4 = (
            2
            * n
            * (n + 1)
            * (j(x_s_2) - x_s_2 * dj(x_s_2))
            / ((n + 2) * (n - 1) * j(x_s_2) + x_s_2**2 * ddj(x_s_2))
        )

        F_n = (frac1 - frac2) / (frac3 - frac4)
        F_n *= 1 / 2 * self.rho_f / self.rho_s * x_s_2**2

        c_n = -(
            (F_n * j(x_f) - x_f * dj(x_f)) / (F_n * h(x_f) - x_f * dh(x_f))
        )

        return c_n

    def c_n_A(self, n: int) -> complex:
        return self.cls.c_n(n) * self.cls.A_in(n)

    def a_n(self, n: int) -> complex:

        k_f = self.cls.k_f
        k_s_1 = self.cls.k_s_l
        k_s_2 = self.cls.k_s_t
        x_f = self.R_0 * k_f
        x_s_1 = self.R_0 * k_s_1
        x_s_2 = self.R_0 * k_s_2
        A_in = self.cls.A_in(n)
        c_n = self.cls.c_n_A(n)
        j = lambda x: Bf.besselj(n, x)
        dj = lambda x: Bf.d1_besselj(n, x)
        ddj = lambda x: Bf.d2_besselj(n, x)
        dh = lambda x: Bf.d1_hankelh1(n, x)

        a_num = (
            x_f
            * (c_n * dh(x_f) + A_in * dj(x_f))
            * ((n**2 + n - 1) * x_s_2 * dj(x_s_2) - j(x_s_2))
        )
        a_den = (
            n * (n + 1) * (-(x_s_1**2) * ddj(x_s_1) + j(x_s_1)) * j(x_s_2)
        )
        a_den += (
            x_s_1
            * dj(x_s_1)
            * (
                (n**2 + n - 1) * x_s_2 * dj(x_s_2)
                - (n**2 + n + 1) * j(x_s_2)
            )
        )

        return a_num / a_den

    def b_n(self, n: int) -> complex:

        k_f = self.cls.k_f
        k_s_1 = self.cls.k_s_l
        k_s_2 = self.cls.k_s_t
        x_f = self.R_0 * k_f
        x_s_1 = self.R_0 * k_s_1
        x_s_2 = self.R_0 * k_s_2
        A_in = self.cls.A_in(n)
        c_n = self.cls.c_n_A(n)
        j = lambda x: Bf.besselj(n, x)
        dj = lambda x: Bf.d1_besselj(n, x)
        ddj = lambda x: Bf.d2_besselj(n, x)
        dh = lambda x: Bf.d1_hankelh1(n, x)

        b_nom = (
            -x_f
            * (c_n * dh(x_f) + A_in * dj(x_f))
            * (x_s_1 * (x_s_1 * ddj(x_s_1) + dj(x_s_1)) - j(x_s_1))
        )
        b_num = n * (n + 1) * (x_s_1**2 * ddj(x_s_1) - j(x_s_1)) * j(
            x_s_2,
        ) + (
            x_s_1
            * dj(x_s_1)
            * (
                -(n**2 + n - 1) * x_s_2 * dj(x_s_2)
                + (n**2 + n + 1) * j(x_s_2)
            )
        )

        return b_nom / b_num

    def potential_coefficient(self, n: int) -> complex:
        return -self.c_n_A(n)

    def V_r_sc(self, n, r):
        c_n = self.cls.c_n_A(n)
        k_f = self.cls.k_f
        return k_f * c_n * Bf.d1_hankelh1(n, k_f * r)

    def V_theta_sc(self, n, r):
        c_n = self.cls.c_n_A(n)
        k_f = self.cls.k_f
        return c_n * Bf.hankelh1(n, k_f * r) / r

    def radial_particle_velocity(self, r, theta, t, mode):
        def radial_func(l, x):
            k_s_1 = self.cls.k_s_l
            k_s_2 = self.cls.k_s_t
            a_n = self.cls.a_n(l)
            b_n = self.cls.b_n(l)
            j = Bf.besselj
            dj = Bf.d1_besselj
            term1 = k_s_1 * a_n * dj(l, k_s_1 * x)
            term2 = l * (l + 1) / x * b_n * j(l, k_s_2 * x)
            return term1 - term2

        return self.cls.radial_mode_superposition(
            radial_func,
            r,
            theta,
            t,
            mode,
        )

    def tangential_particle_velocity(self, r, theta, t, mode):
        def radial_func(l, x):
            k_s_1 = self.cls.k_s_l
            k_s_2 = self.cls.k_s_t
            a_n = self.cls.a_n(l)
            b_n = self.cls.b_n(l)
            j = Bf.besselj
            dj = Bf.d1_besselj
            term1 = a_n * j(l, k_s_1 * x)
            term2 = b_n * (j(l, k_s_2 * x) + k_s_2 * x * dj(l, k_s_2 * x))
            return (term1 - term2) / x

        return self.cls.tangential_mode_superposition(
            radial_func,
            r,
            theta,
            t,
            mode,
        )

    def test_methods_n(self) -> None:
        methods = {}
        methods["c_n_A"] = None
        methods["c_n"] = None
        methods["a_n"] = None
        methods["b_n"] = None
        methods["potential_coefficient"] = None
        self._test_methods_n(methods, n_end=4, threshold=1e-5)

    def test_V_r_sc(self):
        for n in range(self.n_runs):
            r = self.get_random_r(self.cls)
            self.do_testing(
                func_1=self.cls.V_r_sc,
                args_1=(n, r),
                func_2=self.V_r_sc,
                args_2=(n, r),
            )

    def test_V_theta_sc(self):
        for n in range(self.n_runs):
            r = self.get_random_r(self.cls)
            self.do_testing(
                func_1=self.cls.V_theta_sc,
                args_1=(n, r),
                func_2=self.V_theta_sc,
                args_2=(n, r),
            )


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from osaft import log
from osaft.core.backgroundfields import WaveType
from osaft.core.frequency import Frequency
from osaft.core.geometries import Sphere
from osaft.core.helper import StringFormatter as SF
from osaft.core.solids import RigidSolid
from osaft.core.variable import ActiveVariable
from osaft.solutions.base_solution import BaseSolution
from osaft.solutions.basedoinikov1994.base import BaseDoinikov1994


class BaseDoinikov1994Rigid(BaseDoinikov1994, BaseSolution):
    """Base class for Doinikov (viscous fluid-rigid sphere; 1994)

    :param f: Frequency [Hz]
    :param R_0: Radius of the sphere [m]
    :param rho_s: Density of the sphere [kg/m^3]
    :param rho_f: Density of the fluid [kg/m^3]
    :param c_f: Speed of sound of the fluid [m/s]
    :param eta_f: shear viscosity [Pa s]
    :param zeta_f: bulk viscosity [Pa s]
    :param p_0: Pressure amplitude of the field [Pa]
    :param wave_type: Type of wave, traveling or standing
    :param position: Position in the standing wave field [rad]
    """

    # Supported wave_type for the class
    supported_wavetypes = [WaveType.STANDING, WaveType.TRAVELLING]

    def __init__(
        self,
        f: Frequency | float,
        R_0: Sphere | float | int,
        rho_s: float,
        rho_f: float,
        c_f: float,
        eta_f: float,
        zeta_f: float,
        p_0: float,
        wave_type: None | WaveType = WaveType.STANDING,
        position: None | float = None,
    ) -> None:
        """Constructor method"""

        # init of parent class
        BaseDoinikov1994.__init__(
            self,
            f=f,
            R_0=R_0,
            rho_f=rho_f,
            c_f=c_f,
            eta_f=eta_f,
            zeta_f=zeta_f,
            p_0=p_0,
            wave_type=wave_type,
            position=position,
        )
        BaseSolution.__init__(self, "Doinikov1994Rigid")

        # Scatterer
        self.scatterer = RigidSolid(self.f, rho_s)

        # Dependent Variables
        self._rho_t = ActiveVariable(
            self._compute_rho_t,
            "ratio of densities",
        )
        self._rho_t.is_computed_by(self.fluid._rho_f, self.scatterer._rho_s)

        if type(self) == BaseDoinikov1994Rigid:
            log.info(repr(self))  # pragma: no cover

    def __repr__(self):  # pragma: no cover
        return (
            f"BaseDoinikov1994Rigid(f={self.f}, R_0={self.R_0}, "
            f"rho_s={self.rho_s}, rho_f={self.rho_f}, c_f={self.c_f}, "
            f"eta_f={self.eta_f}, zeta_f={self.zeta_f}, "
            f"p_0={self.p_0}, position={self.position}, {self.wave_type}, "
            f"N_max={self.N_max})"
        )

    def __str__(self):  # pragma: no cover
        out = "Doinikov's  (1994) model (viscous fluid-rigid sphere)"
        out += " with the following properties: \n"
        out += SF.get_str_text("Frequency", "f", self.f, "Hz")
        out += SF.get_str_text("Pressure", "p_0", self.p_0, "Pa")
        out += SF.get_str_text("Wavetype", "wave_type", self.wave_type, None)
        out += SF.get_str_text(
            "Position",
            "d",
            self.position,
            "Pa",
        )
        out += SF.get_str_text(
            "Wavelength",
            "lambda",
            self.field.lambda_f,
            "m",
        )

        out += "Fluid\n"
        out += SF.get_str_text(
            "Density",
            "rho_f",
            self.rho_f,
            "kg/m^3",
        )
        out += SF.get_str_text(
            "Sound Speed",
            "c_0",
            self.c_f,
            "m/s",
        )
        out += SF.get_str_text(
            "Compressibility",
            "kappa_f",
            self.kappa_f,
            "1/Pa",
        )
        out += SF.get_str_text(
            "Shear viscosity",
            "eta_f",
            self.eta_f,
            "1/Pa",
        )
        out += SF.get_str_text(
            "Bulk viscosity",
            "zeta_f",
            self.zeta_f,
            "1/Pa",
        )

        out += "Particle\n"
        out += SF.get_str_text(
            "Radius",
            "R_0",
            self.R_0,
            "m",
        )
        out += SF.get_str_text(
            "Density",
            "rho_s",
            self.rho_s,
            "kg/m^3",
        )
        return out

    # -------------------------------------------------------------------------
    # Wrappers for Independent Scatterer Attributes
    # -------------------------------------------------------------------------

    @property
    def rho_s(self) -> float:
        """Wraps to :attr:`osaft.core.solids.RigidSolid.rho_s`"""
        return self.scatterer.rho_s

    @rho_s.setter
    def rho_s(self, value: float) -> None:
        self.scatterer.rho_s = value


if __name__ == "__main__":
    pass



import unittest
from enum import auto

import numpy as np

from osaft import WaveType
from osaft.core.backgroundfields import WrongWaveTypeError
from osaft.core.functions import pi, sin
from osaft.tests.basetest_arf import HelperCompareARF
from osaft.tests.solution_factory import SolutionFactory
from osaft.tests.solutions.king1934.test_base import BaseKing


class TestARF(BaseKing):
    def setUp(self) -> None:
        super().setUp()

        self.parameters._N_max.low = 1
        self.parameters.N_max = 1

        self.cls = SolutionFactory().king_1934_arf()
        self.list_cls = [self.cls]

    # -------------------------------------------------------------------------
    # Getters for Dependent Variables
    # -------------------------------------------------------------------------

    @property
    def f_2(self):
        rho_tilde = self.rho_s / self.rho_f
        return 2 * (rho_tilde - 1) / (2 * rho_tilde + 1)

    @property
    def Phi(self):
        if self.wave_type == WaveType.TRAVELLING:
            return self.compute_Phi_traveling()
        elif self.wave_type == WaveType.STANDING:
            return self.compute_Phi_standing()

    def compute_Phi_standing(self):
        return 1 / 3 + self.f_2 / 2

    def compute_Phi_traveling(self):
        rho_tilde = self.rho_s / self.rho_f

        out = 1 + 2 / 9 * (1 - 1 / rho_tilde) ** 2
        out /= (2 + 1 / rho_tilde) ** 2
        return out

    def FFGG(self, n):
        alpha = self.cls.alpha
        out = self.F_n(n, alpha) * self.F_n(n + 1, alpha)
        out += self.G_n(n, alpha) * self.G_n(n + 1, alpha)
        return out

    def FGFG(self, n):
        alpha = self.cls.alpha
        out = self.F_n(n + 1, alpha)
        out *= self.G_n(n, alpha)
        out -= self.G_n(n + 1, alpha) * self.F_n(n, alpha)
        return out

    def HH(self, n):
        out = self.cls.alpha**2
        if n == 1:
            out -= 3 * (1 - 1 / self.cls.rho_tilde)
        else:
            out -= n * (n + 2)
        out = out**2
        out /= self.cls.alpha ** (4 * n + 6)
        out += self.FFGG(n) ** 2
        return out

    def compute_general_ARF(self):
        if self.wave_type == WaveType.TRAVELLING:
            return self.compute_traveling_general_ARF()
        elif self.wave_type == WaveType.STANDING:
            return self.compute_standing_general_ARF()

    def compute_standing_general_ARF(self):
        alpha = self.cls.alpha
        # n = 0
        out = self.FFGG(0)
        out /= self.HH(0)
        out /= alpha
        # n = 1
        tmp = 2 / alpha**5
        tmp *= self.FFGG(1)
        tmp /= self.HH(1)
        tmp *= alpha**2 - 3 * (1 - 1 / self.cls.rho_tilde)
        out -= tmp
        # n > 1
        for i in range(2, self.cls.N_max):
            tmp = (-1) ** i
            tmp *= (i + 1) / alpha ** (2 * i + 3)
            tmp *= self.FFGG(i)
            tmp /= self.HH(i)
            tmp *= alpha**2 - i * (i + 2)

            out += tmp

        out *= pi * self.cls.rho_f
        out *= np.abs(self.cls.field.A) ** 2
        out *= sin(2 * self.position)

        return out

    def compute_traveling_general_ARF(self):
        alpha = self.cls.alpha
        # n = 0
        out = 1 / self.HH(0)
        # n = 1
        tmp = 2 / self.HH(1)
        tmp *= (alpha**2 - 3 * (1 - 1 / self.cls.rho_tilde)) ** 2
        tmp /= alpha**8
        out += tmp
        # n > 1
        for i in range(2, self.cls.N_max):
            tmp = (i + 1) / self.HH(i)
            tmp *= (alpha**2 - i * (i + 2)) ** 2
            tmp /= alpha ** (4 * i + 4)
            out += tmp

        out *= 2 * pi * self.cls.rho_f
        out *= np.abs(self.cls.field.A) ** 2 / alpha**2

        return out

    # -------------------------------------------------------------------------
    # Tests
    # -------------------------------------------------------------------------

    def test_properties(self) -> None:
        properties = ["Phi", "f_2"]
        self._test_properties(properties)

    def test_methods_n(self) -> None:
        dict_method = {}
        dict_method["FGFG"] = None
        dict_method["FFGG"] = None
        dict_method["HH"] = None

        self._test_methods_n(dict_method)

    # -------------------------------------------------------------------------
    # Test ARF
    # -------------------------------------------------------------------------

    def test_general_solution(self) -> None:
        for n in range(0, self.n_runs + 3):
            self.do_testing(
                func_1=self.cls.compute_arf,
                func_2=self.compute_general_ARF,
                threshold=5e-4,
            )

    def test_NotImplemented_solution(self) -> None:
        WaveType.WRONGWAVETYPE = auto()
        self.cls.wave_type = WaveType.WRONGWAVETYPE
        self.assertRaises(WrongWaveTypeError, self.cls.compute_arf)

    def test_not_small_particle(self):
        # This method is not actually testing, only for coverage
        self.cls.R_0 = self.cls.fluid.lambda_f
        self.cls.small_particle_limit = True
        self.cls.compute_arf()


class TestCompareToGorkov(BaseKing, HelperCompareARF):
    def setUp(self) -> None:
        super().setUp()

        self.parameters.R_0 = 1e-7
        self.parameters._R_0.high = 1e-6
        self.parameters.c_s = 1e10
        self.parameters._c_s.low = 1e9
        self.arf_compare_threshold = 1e-4

        self.cls = SolutionFactory().king_1934_arf()
        self.cls.small_particle_limit = False
        self.compare_cls = SolutionFactory().gorkov_1962_arf()
        self.list_cls = [self.cls, self.compare_cls]
        self.assign_parameters()


class TestCompareToKing(HelperCompareARF, BaseKing):
    def setUp(self) -> None:
        BaseKing.setUp(self)

        self.parameters.R_0 = 1e-7
        self.parameters._R_0.high = 1e-6

        self.arf_compare_threshold = 1e-4

        self.small_particle_limit = True

        self.cls = SolutionFactory().king_1934_arf()
        self.compare_cls = SolutionFactory().king_1934_arf()
        self.compare_cls.small_particle_limit = True

        self.list_cls = [self.cls, self.compare_cls]


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
from matplotlib import pyplot as plt

from osaft.plotting.grid import PolarFluidGrid, PolarGrid, PolarScattererGrid
from osaft.tests.plotting.grid.test_base_grid import plot_grid


class TestPolarGrid(unittest.TestCase):
    def setUp(self) -> None:
        self.r_min = 1
        self.r_max = 2
        self.res_1 = 10
        self.res_2 = 20

        # Grid with the same resolution in both dimensions
        self.grid_1_res = PolarGrid(
            self.r_min,
            self.r_max,
            self.res_1,
            False,
        )
        # Grid with the same resolution in both dimensions from a tuple
        self.grid_1_from_seq_res = PolarGrid(
            self.r_min,
            self.r_max,
            [self.res_1],
            False,
        )
        # Grid with the different resolutions in the two dimensions
        self.grid_2_res = PolarGrid(
            self.r_min,
            self.r_max,
            (self.res_1, self.res_2),
            False,
        )

        # Grid with the same resolution in both dimensions
        self.grid_1_res_upper = PolarGrid(
            self.r_min,
            self.r_max,
            self.res_1,
            True,
        )
        # Grid with the same resolution in both dimensions from a tuple
        self.grid_1_from_seq_res_upper = PolarGrid(
            self.r_min,
            self.r_max,
            [self.res_1],
            True,
        )
        # Grid with the different resolutions in the two dimensions
        self.grid_2_res_upper = PolarGrid(
            self.r_min,
            self.r_max,
            (self.res_1, self.res_2),
            True,
        )

    # -------------------------------------------------------------------------
    # Visual tests (skipped)
    # -------------------------------------------------------------------------

    @unittest.skip("Used only for visual testing during development.")
    def test_plot(self):
        plot_grid(self.grid_1_res)
        plot_grid(self.grid_2_res_upper)
        plot_grid(self.grid_1_res)
        plot_grid(self.grid_2_res_upper)
        plt.show()

    # -------------------------------------------------------------------------
    # Test step size tests
    # -------------------------------------------------------------------------

    def test_dr(self):
        dr = (self.r_max - self.r_min) / (self.res_1 - 1)
        self.assertAlmostEqual(self.grid_1_res.dr, dr)
        self.assertAlmostEqual(self.grid_1_from_seq_res.dr, dr)
        self.assertAlmostEqual(self.grid_2_res.dr, dr)

    def test_dtheta(self):
        dtheta_1 = 2 * np.pi / (self.res_1 - 1)
        dtheta_2 = 2 * np.pi / (self.res_2 - 1)
        self.assertAlmostEqual(self.grid_1_res.dtheta, dtheta_1)
        self.assertAlmostEqual(self.grid_1_from_seq_res.dtheta, dtheta_1)
        self.assertAlmostEqual(self.grid_2_res.dtheta, dtheta_2)

        dtheta_1_upper = np.pi / (self.res_1 // 2 - 1)
        dtheta_2_upper = np.pi / (self.res_2 - 1)
        self.assertEqual(self.grid_1_res_upper.dtheta, dtheta_1_upper)
        self.assertEqual(self.grid_1_from_seq_res_upper.dtheta, dtheta_1_upper)
        self.assertEqual(self.grid_2_res_upper.dtheta, dtheta_2_upper)

    # -------------------------------------------------------------------------
    # Array / resolution tests
    # -------------------------------------------------------------------------

    def test_arr_r(self) -> None:
        arr_r = np.linspace(self.r_min, self.r_max, self.res_1)
        np.testing.assert_array_equal(self.grid_1_res.arr_r, arr_r)
        np.testing.assert_array_equal(self.grid_1_from_seq_res.arr_r, arr_r)
        np.testing.assert_array_equal(self.grid_2_res.arr_r, arr_r)

    def test_arr_theta(self) -> None:
        arr_theta_1 = np.linspace(-np.pi, np.pi, self.res_1)
        arr_theta_2 = np.linspace(-np.pi, np.pi, self.res_2)
        np.testing.assert_array_equal(self.grid_1_res.arr_theta, arr_theta_1)
        np.testing.assert_array_equal(
            self.grid_1_from_seq_res.arr_theta,
            arr_theta_1,
        )
        np.testing.assert_array_equal(self.grid_2_res.arr_theta, arr_theta_2)

        arr_theta_1_upper = np.linspace(0, np.pi, self.res_1 // 2)
        arr_theta_2_upper = np.linspace(0, np.pi, self.res_2)
        np.testing.assert_array_equal(
            self.grid_1_res_upper.arr_theta,
            arr_theta_1_upper,
        )
        np.testing.assert_array_equal(
            self.grid_1_from_seq_res_upper.arr_theta,
            arr_theta_1_upper,
        )
        np.testing.assert_array_equal(
            self.grid_2_res_upper.arr_theta,
            arr_theta_2_upper,
        )

    # -------------------------------------------------------------------------
    # Test meshes
    # -------------------------------------------------------------------------

    def test_meshes(self) -> None:

        # Upper = False
        arr_r = self.grid_1_res.arr_r
        arr_t = self.grid_1_res.arr_theta
        mesh_t, mesh_r = np.meshgrid(arr_t, arr_r)

        np.testing.assert_array_equal(mesh_r, self.grid_1_res.mesh_r)
        np.testing.assert_array_equal(mesh_t, self.grid_1_res.mesh_theta)

        # Upper = True
        arr_r = self.grid_1_res_upper.arr_r
        arr_t = self.grid_1_res_upper.arr_theta
        mesh_t, mesh_r = np.meshgrid(arr_t, arr_r)

        np.testing.assert_array_equal(mesh_r, self.grid_1_res_upper.mesh_r)
        np.testing.assert_array_equal(mesh_t, self.grid_1_res_upper.mesh_theta)

    # -------------------------------------------------------------------------
    # Test conversions
    # -------------------------------------------------------------------------

    def test_conversion(self) -> None:

        # Upper = False
        mesh_r = self.grid_1_res.mesh_r
        mesh_t = self.grid_1_res.mesh_theta

        mesh_x = mesh_r * np.sin(mesh_t)
        mesh_z = mesh_r * np.cos(mesh_t)

        np.testing.assert_array_equal(mesh_x, self.grid_1_res.mesh_x)
        np.testing.assert_array_equal(mesh_z, self.grid_1_res.mesh_z)

        # Upper = True
        mesh_r = self.grid_1_res_upper.mesh_r
        mesh_t = self.grid_1_res_upper.mesh_theta

        mesh_x = mesh_r * np.sin(mesh_t)
        mesh_z = mesh_r * np.cos(mesh_t)

        np.testing.assert_array_equal(mesh_x, self.grid_1_res_upper.mesh_x)
        np.testing.assert_array_equal(mesh_z, self.grid_1_res_upper.mesh_z)

    def test_polar_scatterer_grid(self) -> None:

        self.scatterer_grid = PolarScattererGrid(1, 10, True)
        self.assertEqual(np.max(self.scatterer_grid.arr_r.min()), 1e-30)
        self.assertEqual(np.max(self.scatterer_grid.arr_r.max()), 1)

    def test_polar_fluid_grid(self) -> None:

        self.scatterer_grid = PolarFluidGrid(1, 2, 10, True)
        self.assertEqual(np.max(self.scatterer_grid.arr_r.min()), 1)
        self.assertEqual(np.max(self.scatterer_grid.arr_r.max()), 2)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from collections.abc import Callable, Iterable

import numpy as np
from scipy import special
from scipy.integrate import quad
from scipy.special import exp1, factorial, lpmv

from osaft import log

# Often used NumPy and SciPy Functions and Classes
NDArray = np.ndarray
conj = np.conj
exp = np.exp
real = np.real
imag = np.imag
sqrt = np.sqrt
sin = np.sin
cos = np.cos
pi = np.pi


def full_range(*args: int) -> Iterable:
    """Returns the range including the end point.

    :param args: end point, start point (optional)
    :raises: ValueError if more than 2 arguments are passed"""
    if len(args) == 1:
        return range(args[0] + 1)
    elif len(args) == 2:
        return range(args[0], args[1] + 1)
    else:
        raise ValueError("full_range takes one or two arguments.")


def cartesian_2_spherical_coordinates(
    x: float | NDArray,
    z: float | NDArray,
) -> tuple[float, float]:
    """Transforms Cartesian coordinates `(x  z)` of an axisymmetric model into
    spherical coordinates `(r, theta)`. See `here
    <https://gitlab.ethz.ch/goeringc/src-git/-/wikis/Definition-of-the-Coordinate-System>`__
    for the definition of the coordinate system.

    :param x: x coordinate
    :param z: z coordinate
    """
    r = np.hypot(x, z)
    theta = np.arctan2(x, z)
    return r, theta


def spherical_2_cartesian_coordinates(
    r: float | NDArray,
    theta: float | NDArray,
) -> tuple[float, float]:
    """Transforms spherical coordinates `(r  theta)` of an axisymmetric model
    into Cartesian coordinates `(x, z)`. See `here
    <https://gitlab.ethz.ch/goeringc/src-git/-/wikis/Definition-of-the-Coordinate-System>`__
    for the definition of the coordinate system.

    :param r: x radial coordinate
    :param theta: polar angle
    """
    x = r * sin(theta)
    z = r * cos(theta)
    return x, z


def transform_vector(v_x: float, v_y: float, theta: float) -> [float, float]:
    """Transforms the coordinates of a vector in the xy-plane from the
    reference system to a coordinate system rotated by `theta`

    :param v_x: first coordinate of the vector
    :param v_y: second coordinate of the vector
    :param theta: rotation angle
    """
    v_xt = sin(theta) * v_x + cos(theta) * v_y
    v_yt = cos(theta) * v_x - sin(theta) * v_y
    return v_xt, v_yt


def spherical_2_cartesian_vector(
    v_r: float,
    v_theta: float,
    theta: float,
) -> tuple[float, float]:
    """Transforms velocities in the Spherical coordinate system
    `(v_r, v_theta)` of an axisymmetric model to the Cartesian system
    `(v_x, v_z)`. See `here
    <https://gitlab.ethz.ch/goeringc/src-git/-/wikis/Definition-of-the-Coordinate-System>`__
    for the definition of the coordinate system.

    :param v_r: velocity in radial direction
    :param v_theta: velocity in tangential direction
    :param theta: polar angle
    """
    return transform_vector(v_r, v_theta, theta)


def cartesian_2_spherical_vector(
    v_x: float,
    v_z: float,
    theta: float,
) -> tuple[float, float]:
    """Transforms velocities in the a Spherical coordinate system
    `(v_r, v_theta)` of an axisymmetric model to the Cartesian system
    `(v_x, v_z)`. See `here
    <https://gitlab.ethz.ch/goeringc/src-git/-/wikis/Definition-of-the-Coordinate-System>`__
    for the definition of the coordinate system.

    :param v_x: velocity in radial direction
    :param v_z: velocity in tangential direction
    :param theta: polar angle
    """
    return transform_vector(v_x, v_z, theta)


def clebsch_gordan_coefficient(
    j1: int,
    m1: int,
    j2: int,
    m2: int,
    j: int,
    m: int,
) -> float:
    """Clebsch-Gordan coefficient for integer valued arguments.

    :param j:
    :param m:
    :param j1:
    :param m1:
    :param j2:
    :param m2:
    """

    for arg in [j1, m1, j2, m2, j, m]:
        if not isinstance(arg, int):
            raise ValueError("Arguments need to be integer valued.")

    if (m1 + m2) != m:
        return 0
    elif j1 + j2 < j or abs(j1 - j2) > j:
        return 0
    else:
        c = 0.0
        z = 0
        a = (
            factorial(j1 + j2 - j)
            * factorial(j1 - j2 + j)
            * factorial(-j1 + j2 + j)
            / factorial(j1 + j2 + j + 1.0)
        ) ** 0.5
        b = (
            factorial(j1 + m1)
            * factorial(j1 - m1)
            * factorial(j2 + m2)
            * factorial(j2 - m2)
            * factorial(j + m)
            * factorial(j - m)
        ) ** 0.5
        while z < (j1 - m1 + 3):
            denominator = (
                factorial(z)
                * factorial(j1 + j2 - j - z)
                * factorial(j1 - m1 - z)
                * factorial(j2 + m2 - z)
                * factorial(j - j2 + m1 + z)
                * factorial(j - j1 - m2 + z)
            )
            numerator = (-1) ** float(z + j1 - j2 + m)
            if denominator != 0:
                c += numerator / denominator
            z += 1
        return (-1.0) ** (j1 - j2 + m) * (2.0 * j + 1.0) ** 0.5 * a * b * c


def integrate(
    func: Callable[[float], float],
    lower: float,
    upper: float,
    rel_eps: float = 1e-6,
) -> tuple[float, float]:
    """Integrates :attr:`func` between :attr:`lower` and :attr:`upper`
    Scipy's QUADPACK implementation `quad()
    <https://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate
    .quad.html>`__ .

    :param func: function to be integrated.
    :param lower: lower integration limit
    :param upper: upper integration limit
    :param rel_eps: estimated relative error of integration
    """

    # Logging
    log.debug(
        "Relative precision goal for integration has been passed."
        "Adaptive integration scheme is chosen.",
    )
    # Integration
    integral, abs_err = quad(
        func,
        lower,
        upper,
        epsabs=0,
        epsrel=rel_eps,
        limit=200,
    )
    log.debug(
        f"Integral = {integral}," f"absolute error estimation = {abs_err}\n",
    )
    return integral, abs_err


def integrate_osc(
    func: Callable[[float], float],
    lower: float,
    upper: float,
    boundary_layer: float,
    viscous_wavelength: None | float = None,
    resolution: int = 4,
    roi_factor: int = 5,
    rel_eps: float = 1e-6,
) -> tuple[float, float]:
    """Special purpose integrator for oscillating integrals that
    integrates :attr:`func` between :attr:`lower` and  :attr:`upper`. This
    method employs Scipy's QUADPACK implementation `quad()
    <https://docs.scipy.org/doc/scipy/reference/generated/
    scipy.integrate.quad.html>`__. Used primarily for boundary layer
    integration for elastic fluids. The method creates a list of
    uniformly distributes points in a region of interest where local
    difficulties are expected.
    The density of points per viscous wavelength is given by
    :attr:`resolution`, the region of interest is given by
    [:attr:`lower`, :attr:`lower` + :attr:`roi_factor` :math:`*`
    :attr:`boundary_layer`].
    Function returns the integral and an error estimate.

    :param func: function to be integrated. If no rel_eps is passed the
        function needs to be vectorized.
    :param lower: lower integration limit
    :param upper: upper integration limit
    :param viscous_wavelength: viscous wavelength in the fluid [m]
    :param boundary_layer: boundary layer thickness in the fluid [m]
    :param resolution: number of integration points per wavelength passed
    :param roi_factor: multiplier to compute ROI  (see above)
    :param rel_eps: estimated relative error of integration
    """

    if viscous_wavelength is None:
        viscous_wavelength = 2 * pi * boundary_layer

    # Generate points
    dx = viscous_wavelength / resolution
    region_of_interest = roi_factor * boundary_layer
    n = int(region_of_interest / dx)
    points = np.linspace(lower, lower + region_of_interest, n)

    # Test if subdivision limit is large enough
    limit = 10 * n if 10 * n > 50 else 50

    # Integrate
    integral, abs_err = quad(
        func,
        lower,
        upper,
        epsabs=0,
        epsrel=rel_eps,
        points=points,
        limit=limit,
    )

    log.debug(
        f"Integral = {integral}," f"absolute error estimation = {abs_err}\n",
    )
    return integral, abs_err


def xexpexp1(z: complex, N: int = 20) -> complex:
    """Computes :math:`z\\exp(z)E_1(z)`

    The function is meant to return accurate values also if
    :math:`\\mathrm{Re}(z)` is large. If :math:`\\mathrm{Re}(z)` is large
    then the divergent series is used

    :param z: argument
    :param N: number of terms to be considered
    """
    if z.real > 20:
        s = 0
        for n in range(N, 0, -1):
            s = (s + (-1) ** n) * n / z
        return s + 1
    else:
        return z * exp(z) * exp1(z)


class LegendreFunctions:
    """This class gathers all reoccurring Legendre Functions necessary for
    the osaft package."""

    @staticmethod
    def cos_poly(theta: float, coefficients: NDArray) -> NDArray:
        """Sum of Legendre polynomial with a cosine function as an argument

        :param theta: angle
        :param coefficients: coefficients of the polynomial
        """
        return np.polynomial.legendre.legval(
            cos(theta),
            coefficients,
            tensor=False,
        )

    @staticmethod
    def cos_monomial(degree: int, theta: float, coefficient: complex) -> float:
        """Returns a single term of Legendre Cosine series

        :param degree: degree of the polynomial to be returned
        :param theta:  angle
        :param coefficient: coefficients of the polynomial
        """
        return coefficient * lpmv(0, degree, cos(theta))

    @staticmethod
    def first_cos_poly(theta: float, coefficients: NDArray) -> NDArray:
        """Sum of associate Legendre Polynomial of order one with a cosine
        as input variable

        :param theta: angle
        :param coefficients: coefficients of the polynomial
        """
        degrees = len(coefficients)
        return np.sum(
            [coefficients[n] * lpmv(1, n, cos(theta)) for n in range(degrees)],
            axis=0,
        )

    @staticmethod
    def first_cos_monomial(
        degree: int,
        theta: float,
        coefficient: complex,
    ) -> float:
        """Returns a single term of associate Legendre Cosine series of the
        first order

        :param degree: degree of the term
        :param theta:  angle
        :param coefficient: coefficient of the polynomial
        """
        return coefficient * lpmv(1, degree, cos(theta))


class BesselFunctions:
    """This class gathers all reoccurring spherical hankel and bessel
    functions necessary for the osaft package."""

    @staticmethod
    def _a0_d2(n: int, z: complex) -> complex:
        return (n**2 - n - z**2) / z**2

    @staticmethod
    def _a1_d2(n: int, z: complex) -> complex:
        return 2 / z

    @staticmethod
    def _a0_d3(n: int, z: complex) -> complex:
        return (n**3 - 3 * n**2 + 2 * n) / z**3 + (2 - n) / z

    @staticmethod
    def _a1_d3(n: int, z: complex) -> complex:
        return (-(n**2) - n - 6) / z**2 + 1

    @staticmethod
    def _a0_d4(n: int, z: complex) -> complex:
        return (
            n * (n - 1) * (6 + n * (n - 5)) / z**4
            + (-2 * n * (n - 1) - 8) / z**2
            + 1
        )

    @staticmethod
    def _a1_d4(n: int, z: complex) -> complex:
        return 8 * (3 + n + n**2) / z**3 - 4 / z

    @staticmethod
    def _a0_d5(n: int, z: complex) -> complex:
        return (
            (-4 + n) * (-3 + n) * (-2 + n) * (-1 + n) * n
            - 2 * (-20 + n * (2 + (-7 + n) * n)) * z**2
            + (-4 + n) * z**4
        ) / z**5

    @staticmethod
    def _a1_d5(n: int, z: complex) -> complex:
        return -(
            (
                n * (1 + n) * (58 + n + n**2)
                - 2 * n * (1 + n) * z**2
                + z**4
                - 20 * (-6 + z**2)
            )
            / z**4
        )

    @staticmethod
    def besselj(n: int, z: complex) -> complex:
        """Spherical Bessel function of the first kind of order `n`

        :param n: Order
        :param z: argument
        """
        return special.spherical_jn(n, z)

    @classmethod
    def d1_besselj(cls, n: int, z: complex) -> complex:
        """First derivative of the spherical Bessel function of the first kind
        of order `n`

        :param n: order
        :param z: argument
        """
        return special.spherical_jn(n, z, derivative=True)

    @classmethod
    def d2_besselj(cls, n: int, z: complex) -> complex:
        """Second derivative of the spherical Bessel function of the first kind
        of order `n`

        :param n: order
        :param z: argument
        """
        return cls._a0_d2(n, z) * cls.besselj(n, z) + cls._a1_d2(
            n,
            z,
        ) * cls.besselj(n + 1, z)

    @classmethod
    def d3_besselj(cls, n: int, z: complex) -> complex:
        """Third derivative of the spherical Bessel function of the first kind
        of order `n`

        :param n: order
        :param z: argument
        """
        return cls._a0_d3(n, z) * cls.besselj(n, z) + cls._a1_d3(
            n,
            z,
        ) * cls.besselj(n + 1, z)

    @classmethod
    def d4_besselj(cls, n: int, z: complex) -> complex:
        """Fourth derivative of the spherical Bessel function of the first kind
        of order `n`

        :param n: order
        :param z: argument
        """
        return cls._a0_d4(n, z) * cls.besselj(n, z) + cls._a1_d4(
            n,
            z,
        ) * cls.besselj(n + 1, z)

    @staticmethod
    def bessely(n: int, z: complex) -> complex:
        """Spherical Bessel function of the second kind of order `n`

        :param n: order
        :param z: argument
        """
        return special.spherical_yn(n, z)

    @classmethod
    def d1_bessely(cls, n: int, z: complex) -> complex:
        """First derivative of the spherical Bessel function of the second
        kind of order `n`

        :param n: order
        :param z: argument
        """
        return special.spherical_yn(n, z, True)

    @classmethod
    def d2_bessely(cls, n: int, z: complex) -> complex:
        """Second derivative of the spherical Bessel function of the second
        kind of order `n`

        :param n: order
        :param z: argument
        """
        return cls._a0_d2(n, z) * cls.bessely(n, z) + cls._a1_d2(
            n,
            z,
        ) * cls.bessely(n + 1, z)

    @classmethod
    def d3_bessely(cls, n: int, z: complex) -> complex:
        """Third derivative of the spherical Bessel function of the second
        kind of order `n`

        :param n: order
        :param z: argument
        """
        return cls._a0_d3(n, z) * cls.bessely(n, z) + cls._a1_d3(
            n,
            z,
        ) * cls.bessely(n + 1, z)

    @classmethod
    def d4_bessely(cls, n: int, z: complex) -> complex:
        """Fourth derivative of the spherical Bessel function of the second
        kind of order `n`

        :param n: order
        :param z: argument
        """
        return cls._a0_d4(n, z) * cls.bessely(n, z) + cls._a1_d4(
            n,
            z,
        ) * cls.bessely(n + 1, z)

    @staticmethod
    def hankelh1(n: int, z: complex) -> complex:
        """Spherical Hankel function of the first kind of order `n`

        :param n: order
        :param z: argument
        """
        return sqrt(pi / (2 * z)) * special.hankel1(n + 0.5, z)

    @classmethod
    def d1_hankelh1(cls, n: int, z: complex) -> complex:
        """First derivative of the spherical Hankel function of the first
        kind of order `n`

        :param n: order
        :param z: argument
        """
        return -cls.hankelh1(n + 1, z) + (n / z) * cls.hankelh1(n, z)

    @classmethod
    def d2_hankelh1(cls, n: int, z: complex) -> complex:
        """Second derivative of the spherical Hankel function of the first
        kind of order `n`

        :param n: order
        :param z: argument
        """
        return cls._a0_d2(n, z) * cls.hankelh1(n, z) + cls._a1_d2(
            n,
            z,
        ) * cls.hankelh1(n + 1, z)

    @classmethod
    def d3_hankelh1(cls, n: int, z: complex) -> complex:
        """Third derivative of the spherical Hankel function of the first
        kind of order `n`

        :param n: order
        :param z: argument
        """
        return cls._a0_d3(n, z) * cls.hankelh1(n, z) + cls._a1_d3(
            n,
            z,
        ) * cls.hankelh1(n + 1, z)

    @classmethod
    def d4_hankelh1(cls, n: int, z: complex) -> complex:
        """Fourth derivative of the spherical Hankel function of the first
        kind of order `n`

        :param n: order
        :param z: argument
        """
        return cls._a0_d4(n, z) * cls.hankelh1(n, z) + cls._a1_d4(
            n,
            z,
        ) * cls.hankelh1(n + 1, z)

    @classmethod
    def d5_hankelh1(cls, n: int, z: complex) -> complex:
        """Fifth derivative of the spherical Hankel function of the first
        kind of order `n`

        :param n: order
        :param z: argument
        """
        return cls._a0_d5(n, z) * cls.hankelh1(n, z) + cls._a1_d5(
            n,
            z,
        ) * cls.hankelh1(n + 1, z)

    @staticmethod
    def hankelh2(n: int, z: complex) -> complex:
        """Spherical Hankel function of the second kind of order `n`

        :param n: order
        :param z: argument
        """
        return sqrt(pi / (2 * z)) * special.hankel2(n + 0.5, z)

    @classmethod
    def d1_hankelh2(cls, n: int, z: complex) -> complex:
        """First derivative of the spherical Hankel function of the second
        kind of order `n` second kind

        :param n: order
        :param z: argument
        """
        return -cls.hankelh2(n + 1, z) + (n / z) * cls.hankelh2(n, z)

    @classmethod
    def d2_hankelh2(cls, n: int, z: complex) -> complex:
        """Second derivative of the spherical Hankel function of the second
        kind of order `n` second kind

        :param n: order
        :param z: argument
        """
        return cls._a0_d2(n, z) * cls.hankelh2(n, z) + cls._a1_d2(
            n,
            z,
        ) * cls.hankelh2(n + 1, z)

    @classmethod
    def d3_hankelh2(cls, n: int, z: complex) -> complex:
        """Third derivative of the spherical Hankel function of the second
        kind of order `n` second kind

        :param n: order
        :param z: argument
        """
        return cls._a0_d3(n, z) * cls.hankelh2(n, z) + cls._a1_d3(
            n,
            z,
        ) * cls.hankelh2(n + 1, z)

    @classmethod
    def d4_hankelh2(cls, n: int, z: complex) -> complex:
        """Fourth derivative of the spherical Hankel function of the second
        kind of order `n` second kind

        :param n: order
        :param z: argument
        """
        return cls._a0_d4(n, z) * cls.hankelh2(n, z) + cls._a1_d4(
            n,
            z,
        ) * cls.hankelh2(n + 1, z)

    @classmethod
    def d5_hankelh2(cls, n: int, z: complex) -> complex:
        """Fifth derivative of the spherical Hankel function of the second
        kind of order `n` second kind

        :param n: order
        :param z: argument
        """
        return cls._a0_d5(n, z) * cls.hankelh2(n, z) + cls._a1_d5(
            n,
            z,
        ) * cls.hankelh2(n + 1, z)

    @classmethod
    def adaptive_derivative_besselj(
        cls,
        n: int,
        z: complex,
        i: int = 0,
    ) -> complex:
        """Returns the value of the `i`-th derivative of the spherical
        Bessel function of the first kind of order `n`. The
        coefficients for the derivative are calculated at runtime.

        :param n: order of spherical Bessel function
        :param z: argument of the function
        :param i: i-th derivative is computed
        """
        if not i:
            return cls.besselj(n, z)
        return cls._adaptive_derivative(cls.besselj, i, n, z)

    @classmethod
    def adaptive_derivative_bessely(
        cls,
        n: int,
        z: complex,
        i: int = 0,
    ) -> complex:
        """Returns the value of the `i`-th derivative of the spherical
        Bessel function of the second kind of order `n`. The coefficients
        for the derivative are calculated at runtime.

        :param n: order of spherical Hankel function
        :param z: argument of the function
        :param i: i-th derivative is computed at runtime.
        """
        if not i:
            return cls.bessely(n, z)
        return cls._adaptive_derivative(cls.bessely, i, n, z)

    @classmethod
    def adaptive_derivative_hankelh1(
        cls,
        n: int,
        z: complex,
        i: int = 0,
    ) -> complex:
        """Returns the value of the `i`-th derivative of the spherical
        Hankel function of order `n`. The coefficients for the derivative
        are calculated at runtime.

        :param n: order of spherical Hankel function
        :param z: argument of the function
        :param i: i-th derivative is computed
        """
        if not i:
            return cls.hankelh1(n, z)
        return cls._adaptive_derivative(cls.hankelh1, i, n, z)

    @classmethod
    def adaptive_derivative_hankelh2(
        cls,
        n: int,
        z: complex,
        i: int = 0,
    ) -> complex:
        """Returns the value of the `i`-th derivative of the spherical
        Hankel function of order `n`. The coefficients for the derivative
        are calculated at runtime.

        :param n: order of spherical Hankel function
        :param z: argument of the function
        :param i: i-th derivative is computed
        """
        if not i:
            return cls.hankelh2(n, z)
        return cls._adaptive_derivative(cls.hankelh2, i, n, z)

    @classmethod
    def _adaptive_derivative(
        cls,
        func: Callable[[int, complex], complex],
        i: int,
        n: int,
        z: complex,
    ) -> complex:
        out = 0
        for j in np.arange(i + 1):
            out += cls._adaptive_coefficient(
                i,
                j,
                n,
                z,
            ) * func(n + j, z)
        return out

    @classmethod
    def _adaptive_coefficient(
        cls,
        i: int,
        j: int,
        n: int,
        z: complex,
    ) -> complex:
        if j > i or j < 0:
            out = 0
        elif j == i:
            out = (-1) ** i
        elif j == 0 and i == 1:
            out = n / z
        else:
            out = (n + 2 * j + 1 - i) / z
            out *= cls._adaptive_coefficient(i - 1, j, n, z)
            out -= cls._adaptive_coefficient(i - 1, j - 1, n, z)

        return out


if __name__ == "__main__":
    pass



import unittest

from osaft.solutions.base_scattering import BaseScattering
from osaft.tests.plotting.scattering.setup_test_scattering import (
    BaseTestFluidQuiverPlot,
)
from osaft.tests.solution_factory import SolutionFactory


class TestKingFluidScattering(BaseTestFluidQuiverPlot):
    def setUp(self):
        super().setUp()
        self.name_prefix = "King1934_Fluid"

    def _get_solution(self) -> BaseScattering:
        return SolutionFactory().king_1934_scattering()


class TestYosiokaFluidScattering(BaseTestFluidQuiverPlot):
    def setUp(self):
        super().setUp()
        self.name_prefix = "Yosioka_Fluid"

    def _get_solution(self) -> BaseScattering:
        return SolutionFactory().yosioka_1955_scattering()


class TestHasegawaFluidScattering(BaseTestFluidQuiverPlot):
    def setUp(self):
        super().setUp()
        self.name_prefix = "Hasegawa1969_Fluid"

    def _get_solution(self) -> BaseScattering:
        return SolutionFactory().hasegawa_1969_scattering()


class TestDoiniRigidFluidScattering(BaseTestFluidQuiverPlot):
    def setUp(self):
        super().setUp()
        self.name_prefix = "Doini1994Rigid_Fluid"

    def _get_solution(self) -> BaseScattering:
        return SolutionFactory().doinikov_1994_rigid_scattering()


class TestDoiniCompressibleFluidScattering(BaseTestFluidQuiverPlot):
    def setUp(self):
        super().setUp()
        self.name_prefix = "Doini1994Compressible_Fluid"

    def _get_solution(self) -> BaseScattering:
        return SolutionFactory().doinikov_1994_compressible_scattering()


class TestDoiniViscous2021FluidScattering(BaseTestFluidQuiverPlot):
    def setUp(self):
        super().setUp()
        self.name_prefix = "Doini2021Viscous_Fluid"

    def _get_solution(self) -> BaseScattering:
        return SolutionFactory().doinikov_2021_viscous_scattering()


class TestDoiniViscoelastic2021FluidScattering(BaseTestFluidQuiverPlot):
    def setUp(self):
        super().setUp()
        self.name_prefix = "Doini2021Viscoelastic_Fluid"

    def _get_solution(self) -> BaseScattering:
        return SolutionFactory().doinikov_2021_viscoelastic_scattering()


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from osaft import log
from osaft.core.backgroundfields import BackgroundField, WaveType
from osaft.core.basecomposite import BaseSphereFrequencyComposite
from osaft.core.fluids import InviscidFluid, ViscousFluid
from osaft.core.frequency import Frequency
from osaft.core.functions import pi, sin
from osaft.core.geometries import Sphere
from osaft.core.helper import StringFormatter as SF
from osaft.core.variable import ActiveVariable
from osaft.core.warnings import EPS, raise_assumption_warning
from osaft.solutions.base_arf import BaseARF
from osaft.solutions.base_solution import BaseSolution


class ARF(BaseARF, BaseSphereFrequencyComposite, BaseSolution):
    """ARF class for Settnes and Bruus (2012)

    The standing wave solution is based on Eqs. (50a) - (50b) of the
    paper and the
    traveling wave solution on Eq (48).

    .. note::
       This model is based on the following assumptions:
       - :math:`\\lambda \\gg R_0, \\, \\delta`

    .. note::
        The expression for the travelling wave has been corrected. The
        expression in the article by Settnes & Bruus was off by a factor of 2.
        The error was reported in an article
        `Marston (2013) <https://asa.scitation.org/doi/pdf/10.1121/1.4799400>`_

    :param f: Frequency [Hz]
    :param R_0: Radius of the sphere [m]
    :param rho_s: Density of the sphere [kg/m^3]
    :param c_s: Speed of sound of the fluid-like sphere [m/s]
    :param rho_f: Density of the fluid [kg/m^3]
    :param c_f: Speed of sound of the fluid [m/s]
    :param eta_f: Dynamic viscosity of the fluid [Pa s]
    :param p_0: Pressure amplitude of the field [Pa]
    :param wave_type: Type of incident wave (traveling/standing)
    :param position: Position in the standing wave field [rad]
    """

    # Supported wave_type for the class
    supported_wavetypes = [WaveType.STANDING, WaveType.TRAVELLING]

    def __init__(
        self,
        f: Frequency | float | int,
        R_0: Sphere | float | int,
        rho_s: float,
        c_s: float,
        rho_f: float,
        c_f: float,
        eta_f: float,
        p_0: float,
        wave_type: WaveType,
        position: None | float = None,
    ) -> None:
        """Constructor method"""
        # init of parent class
        # init of parent class
        BaseSphereFrequencyComposite.__init__(self, f, R_0)
        BaseSolution.__init__(self, "Settnes2012")

        # Initialize Components
        self.scatterer = InviscidFluid(self.frequency, rho_s, c_s)
        self.fluid = ViscousFluid(self.frequency, rho_f, c_f, eta_f, 0)
        self.field = BackgroundField(self.fluid, p_0, wave_type, position)

        # Dependent variables
        self._f_1 = ActiveVariable(
            self._compute_f_1,
            "monopole scattering coefficient f_1",
        )
        self._f_2 = ActiveVariable(
            self._compute_f_2,
            "dipole scattering coefficient f_2",
        )
        self._gamma = ActiveVariable(
            self._compute_gamma,
            "dimensionless variable gamma",
        )
        self._delta_t = ActiveVariable(
            self._compute_delta_t,
            "dimensionless boundary layer " "thickness",
        )
        self._kappa_t = ActiveVariable(
            self._compute_kappa_t,
            "compressibility ratio kappa tilde",
        )
        self._rho_t = ActiveVariable(
            self._compute_rho_t,
            "density ratio rho tilde",
        )
        self._Phi = ActiveVariable(
            self._compute_Phi,
            "acoustic contrast factor Phi",
        )

        # Dependencies
        self._f_1.is_computed_by(self._kappa_t)
        self._f_2.is_computed_by(self._rho_t, self._gamma)
        self._delta_t.is_computed_by(self.fluid._delta, self.sphere._R_0)
        self._gamma.is_computed_by(self._delta_t)
        self._kappa_t.is_computed_by(
            self.fluid._kappa_f,
            self.scatterer._kappa_f,
        )
        self._rho_t.is_computed_by(self.fluid._rho_f, self.scatterer._rho_f)
        self._Phi.is_computed_by(self._f_1, self._f_2)

        log.info(str(self))
        log.debug(repr(self))

    def compute_arf(self) -> float:
        """Computes the ARF and returns the force in Newton [N].

        Computes ARF according to Eq. (47a) - (47e) for standing wave case
        or according to Eq. (48) for the travelling wave case.
        Checks before computation of assumption of theory small particle
        radius to pressure field wavelength is valid.

        :raises WrongWaveTypeError: if wrong :attr:`wave_type`
        :raises AssumptionWarning: if the used parameters might not be
            valid for the chosen limiting case
        """
        # Checking wave_type
        self.check_wave_type()

        # size limitation in  section IV.D of paper
        test_value = self.R_0 < EPS * self.field.lambda_f
        test_value = test_value and self.delta < EPS * self.field.lambda_f
        raise_assumption_warning(test_value)

        # Cases
        if self.wave_type == WaveType.STANDING:
            return self._standing_wave_solution()
        else:
            return self._traveling_wave_solution()

    def _standing_wave_solution(self) -> float:
        out = 4 * pi * self.E_ac * self.R_0**3 * self.k_f.real
        out *= self.Phi * sin(2 * self.position)
        return out

    def _traveling_wave_solution(self) -> float:
        # This expression has been corrected according to Marston (2013).
        return (
            2 * pi * self.R_0**3 * self.k_f.real * self.f_2.imag * self.E_ac
        )

    def __repr__(self):
        return (
            f"SettnesARF({self.f}, {self.R_0}, {self.rho_s}, "
            f"{self.c_s}, {self.rho_f}, {self.c_f}, {self.eta_f}"
            f"{self.wave_type}, {self.p_0}, {self.position}"
        )

    def __str__(self):
        out = "Settnes' & Bruus' solution with following properties: \n"
        out += SF.get_str_text("Frequency", "f", self.f, "Hz")
        out += SF.get_str_text("Pressure", "p_0", self.p_0, "Pa")
        out += SF.get_str_text(
            "Position",
            "d",
            self.position,
            "rad",
        )
        out += SF.get_str_text(
            "Wave Type",
            "",
            self.wave_type,
            None,
        )
        out += SF.get_str_text(
            "Wavelength",
            "lambda",
            self.field.lambda_f,
            "m",
        )

        out += "Fluid\n"
        out += SF.get_str_text(
            "Density",
            "rho_f",
            self.rho_f,
            "kg/m^3",
        )
        out += SF.get_str_text(
            "Sound Speed",
            "c_0",
            self.c_f,
            "m/s",
        )
        out += SF.get_str_text(
            "Viscosity",
            "eta_f",
            self.eta_f,
            "Pa s",
        )
        out += SF.get_str_text(
            "Compressibility",
            "kappa_f",
            self.kappa_f,
            "1/Pa",
        )

        out += "Particle\n"
        out += SF.get_str_text(
            "Radius",
            "R_0",
            self.R_0,
            "m",
        )
        out += SF.get_str_text(
            "Density",
            "rho_s",
            self.rho_s,
            "kg/m^3",
        )
        out += SF.get_str_text(
            "Speed of Sound",
            "c_s",
            self.scatterer.c_f,
            "m/s",
        )
        out += SF.get_str_text(
            "Compressibility",
            "kappa_s",
            self.kappa_s,
            "1/Pa",
            False,
        )
        return out

    # -------------------------------------------------------------------------
    # Getters for Dependent Variables
    # -------------------------------------------------------------------------

    @property
    def f_1(self) -> float:
        """Monopole scattering coefficient :math:`f_1` [-]

        (Eq. 26)
        """
        return self._f_1.value

    @property
    def f_2(self) -> complex:
        """Dipole scattering coefficient :math:`f_2` [-]

        (Eq. 39)
        """

        return self._f_2.value

    @property
    def delta_t(self) -> float:
        """Normalized boundary layer thickness :math:`\\tilde{\\delta}` [-]

        (Eq. 34)
        """
        return self._delta_t.value

    @property
    def gamma(self) -> complex:
        """Dimensionless variable :math:`\\gamma` [-]

        (Eq. 38)
        """
        return self._gamma.value

    @property
    def kappa_t(self) -> float:
        """Compressibility ratio :math:`\\tilde{\\kappa}` [-]

        (Eq. 26)
        """
        return self._kappa_t.value

    @property
    def rho_t(self) -> float:
        """Density ratio :math:`\\tilde{\\rho}` [-]

        (Eq. 39)
        """
        return self._rho_t.value

    @property
    def Phi(self) -> float:
        """Acoustic contrast factor :math:`\\Phi` [-]"""
        return self._Phi.value

    # -------------------------------------------------------------------------
    # Wrappers for Independent Field Attributes
    # -------------------------------------------------------------------------

    @property
    def position(self) -> float:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.position`
        """
        return self.field.position

    @position.setter
    def position(self, value: float) -> None:
        self.field.position = value

    @property
    def p_0(self) -> float:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.p_0`
        """
        return self.field.p_0

    @p_0.setter
    def p_0(self, value: float) -> None:
        self.field.p_0 = value

    @property
    def wave_type(self) -> WaveType:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.wave_type`
        """
        return self.field.wave_type

    @wave_type.setter
    def wave_type(self, value: WaveType) -> None:
        self.field.wave_type = value

    # -------------------------------------------------------------------------
    # Wrappers for Independent Field Attributes
    # -------------------------------------------------------------------------

    @property
    def E_ac(self) -> float:
        """Wraps to
        :attr:`osaft.core.backgroundfields.BackgroundField.E_ac`
        """
        return self.field.E_ac

    # -------------------------------------------------------------------------
    # Wrappers for Independent Scatterer Attributes
    # -------------------------------------------------------------------------

    @property
    def rho_s(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.InviscidFluid.rho_s`"""
        return self.scatterer.rho_f

    @rho_s.setter
    def rho_s(self, value: float) -> None:
        self.scatterer.rho_f = value

    @property
    def c_s(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.InviscidFluid.c_s`"""
        return self.scatterer.c_f

    @c_s.setter
    def c_s(self, value: float) -> None:
        self.scatterer.c_f = value

    # -------------------------------------------------------------------------
    # Wrappers for Dependent Scatterer Attributes
    # -------------------------------------------------------------------------

    @property
    def kappa_s(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.InviscidFluid.kappa_f`"""
        return self.scatterer.kappa_f

    # -------------------------------------------------------------------------
    # Wrappers for Independent Fluid Attributes
    # -------------------------------------------------------------------------

    @property
    def rho_f(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.ViscousFluid.rho_f`"""
        return self.fluid.rho_f

    @rho_f.setter
    def rho_f(self, value: float) -> None:
        self.fluid.rho_f = value

    @property
    def c_f(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.ViscousFluid.c_f`"""
        return self.fluid.c_f

    @c_f.setter
    def c_f(self, value: float) -> None:
        self.fluid.c_f = value

    @property
    def eta_f(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.ViscousFluid.eta_f`"""
        return self.fluid.eta_f

    @eta_f.setter
    def eta_f(self, value: float) -> None:
        self.fluid.eta_f = value

    # -------------------------------------------------------------------------
    # Wrappers for Dependent Fluid Attributes
    # -------------------------------------------------------------------------

    @property
    def kappa_f(self) -> float:
        """Wraps to :attr:`osaft.core.fluids.ViscousFluid.kappa_f`"""
        return self.fluid.kappa_f

    @property
    def k_f(self) -> complex:
        """Wraps to :attr:`osaft.core.fluids.ViscousFluid.k_f`"""
        return self.fluid.k_f

    @property
    def delta(self):
        """Wraps to :attr:`osaft.core.fluids.ViscousFluid.k_f`"""
        return self.fluid.delta

    # -------------------------------------------------------------------------
    # Dependent Variables Methods
    # -------------------------------------------------------------------------

    def _compute_f_1(self) -> complex:
        return 1 - self.kappa_s / self.kappa_f

    def _compute_f_2(self) -> complex:
        return (
            2
            * (1 - self.gamma)
            * (self.rho_t - 1)
            / (2 * self.rho_t + 1 - 3 * self.gamma)
        )

    def _compute_delta_t(self) -> float:
        return self.delta / self.R_0

    def _compute_gamma(self) -> complex:
        return -3 / 2 * (1 + 1j * (1 + self.delta_t)) * self.delta_t

    def _compute_kappa_t(self) -> float:
        return self.kappa_s / self.kappa_f

    def _compute_rho_t(self) -> float:
        return self.rho_s / self.rho_f

    def _compute_Phi(self) -> float:
        return 1 / 3 * self.f_1 + 1 / 2 * self.f_2.real


if __name__ == "__main__":
    pass



import unittest
from collections.abc import Iterable

import numpy as np

import osaft
from osaft.core.functions import clebsch_gordan_coefficient as cg
from osaft.core.functions import full_range, integrate, sqrt
from osaft.tests.basetest_numeric import BaseTestSolutions
from osaft.tests.solution_factory import SolutionFactory

conj = osaft.core.functions.conj
NDArray = np.ndarray


class TestStreaming(BaseTestSolutions):
    def setUp(self) -> None:
        super().setUp()

        self._v_boundary_conditions = 1e-3
        self.runs = 5

        self.cls = SolutionFactory().doinikov_2021_viscous_streaming()

        self.list_cls = [self.cls]

    # -------------------------------------------------------------------------
    # Helper Methods
    # -------------------------------------------------------------------------

    @property
    def range_N_max(self) -> Iterable:
        return self.cls.range_N_max

    @property
    def range_1_N_max(self) -> Iterable:
        return self.cls.range_1_N_max

    @property
    def range_2_N_max(self) -> Iterable:
        return self.cls.range_2_N_max

    @property
    def omega(self) -> float:
        return self.cls.omega

    @property
    def k_f(self) -> complex:
        return self.cls.k_f

    @property
    def delta(self):
        return self.cls.delta

    @property
    def k_v(self) -> complex:
        return self.cls.k_v

    def V_r(self, n: int, r: float, ac: bool) -> complex:
        sc = not ac
        return self.cls.V_r(n, r, sc, True)

    def d_V_r(self, n: int, r: float, ac: bool) -> complex:
        sc = not ac
        return self.cls.d_V_r(n, r, sc, True)

    def V_theta(self, n: int, r: float, ac: bool) -> complex:
        sc = not ac
        return self.cls.V_theta(n, r, sc, True)

    def d_V_theta(self, n: int, r: float, ac: bool) -> complex:
        sc = not ac
        return self.cls.d_V_theta(n, r, sc, True)

    def d2_V_theta(self, n: int, r: float, ac: bool):
        sc = not ac
        return self.cls.d2_V_theta(n, r, sc, True)

    def phi(self, n: int, r: float, ac: bool) -> complex:
        sc = not ac
        return self.cls.phi(n, r, sc, True)

    def d_phi(self, n: int, r: float, ac: bool) -> complex:
        sc = not ac
        return self.cls.d_phi(n, r, sc, True)

    # -------------------------------------------------------------------------
    # Methods
    # -------------------------------------------------------------------------

    def inf(self):
        if self.cls.inf_type is not None:
            raise ValueError
        if self.R_0 > self.delta:
            return self.R_0 * self.cls.inf_factor
        else:
            return self.delta * self.cls.inf_factor

    def alpha(self, l: int, r: float, ac: bool) -> float:
        first_sum = complex(0)
        for n in self.range_N_max:
            for m in full_range(abs(l - n), n + l):
                if cg_coef := cg(l, 0, n, 0, m, 0):
                    first_sum += (
                        cg_coef**2
                        / (2 * m + 1)
                        * (
                            self.F(n, m, r, ac)
                            - m * (m + 1) * self.G(n, m, r, ac)
                        )
                    )
        second_sum = complex(0)
        for n in self.range_1_N_max:
            for m in full_range(1, int((n + 1) / 2)):
                for k in full_range(
                    abs(l - (n - 2 * m + 1)),
                    l + n - 2 * m + 1,
                ):
                    if cg_coef := cg(l, 0, n - 2 * m + 1, 0, k, 0):
                        factor = (
                            (2 * n - 4 * m + 3) * cg_coef**2 / (2 * k + 1)
                        )
                        first_term = (
                            (k + 1)
                            * (k + 2)
                            / (2 * k + 3)
                            * self.G(n, k + 1, r, ac)
                        )
                        if k <= 1:
                            second_sum += factor * first_term
                            continue
                        second_term = (
                            k * (k - 1) / (2 * k - 1) * self.G(n, k - 1, r, ac)
                        )
                        second_sum += factor * (first_term - second_term)
        return (2 * l + 1) / (2 * self.omega) * (first_sum + second_sum).real

    def beta(self, l: int, r: float, ac: bool) -> float:
        first_sum = 0
        for n in self.range_N_max:
            first_inner_sum = complex(0)
            for m in full_range(abs(l - n), l + n):
                if cg_coef := cg(l, 0, n, 0, m, 0):
                    first_inner_sum += (
                        cg_coef**2
                        / (2 * m + 1)
                        * conj(
                            self.d_V_r(m, r, ac)
                            - self.k_f**2 * self.phi(m, r, ac),
                        )
                    )
            first_sum += (self.V_r(n, r, ac) * first_inner_sum).real
        second_sum = 0
        for n in self.range_1_N_max:
            for m in full_range(1, int((n + 1) / 2)):
                second_inner_sum = complex(0)
                for k in full_range(
                    abs(l - (n - 2 * m + 1)),
                    l + n - 2 * m + 1,
                ):
                    if cg_coef := cg(l, 0, n - 2 * m + 1, 0, k, 0):
                        factor = cg_coef**2 / (2 * k + 1)
                        first_term = (
                            (k + 1)
                            * (k + 2)
                            / (2 * k + 3)
                            * conj(
                                self.V_r(k + 1, r, ac)
                                - self.V_theta(k + 1, r, ac),
                            )
                        )
                        if k <= 1:
                            second_inner_sum += factor * first_term
                            continue
                        second_term = (
                            k
                            * (k - 1)
                            / (2 * k - 1)
                            * conj(
                                self.V_r(k - 1, r, ac)
                                - self.V_theta(k - 1, r, ac),
                            )
                        )
                        second_inner_sum += factor * (first_term - second_term)
                second_sum += (
                    (2 * n - 4 * m + 3)
                    * self.V_theta(n, r, ac)
                    / r
                    * second_inner_sum
                ).real
        return (2 * l + 1) / 2 * (first_sum + second_sum)

    def gamma(self, l: int, r: float, ac: bool) -> float:
        if l == 0:
            return 0
        first_sum = complex(0)
        for n in self.range_N_max:
            for m in full_range(abs(n - l), n + l):
                cg_coef_1 = cg(n, 0, l, 0, m, 0)
                cg_coef_2 = cg(n, 0, l, 1, m, 1)
                if m >= 1 and cg_coef_1 and cg_coef_2:
                    first_sum += (
                        sqrt(m * (m + 1))
                        * cg_coef_1
                        * cg_coef_2
                        / (2 * m + 1)
                        * (
                            self.V_r(n, r, ac) * conj(self.d_V_theta(m, r, ac))
                            + self.V_theta(m, r, ac)
                            * conj(
                                (
                                    self.V_r(n, r, ac)
                                    - n**2 * self.V_theta(n, r, ac)
                                )
                                / r
                                - self.k_f**2 * self.phi(n, r, ac),
                            )
                        )
                    )
        second_sum = self._V_S_theta_second_sum(l, r, ac)
        return (
            (2 * l + 1)
            / (2 * sqrt(l * (l + 1)))
            * (first_sum + second_sum).real
        )

    def d_gamma(self, l: int, r: float, ac: bool) -> float:
        if l == 0:
            return 0

        first_sum = complex(0)
        for n in self.range_N_max:
            for m in full_range(abs(n - l), n + l):
                cg_coef_1 = cg(n, 0, l, 0, m, 0)
                cg_coef_2 = cg(n, 0, l, 1, m, 1)
                if m >= 1 and cg_coef_1 and cg_coef_2:
                    first_sum += (
                        sqrt(m * (m + 1))
                        * cg_coef_1
                        * cg_coef_2
                        / (2 * m + 1)
                        * (
                            self.d_V_r(n, r, ac)
                            * conj(self.d_V_theta(m, r, ac))
                            + self.V_r(n, r, ac)
                            * conj(self.d2_V_theta(m, r, ac))
                            + self.d_V_theta(m, r, ac)
                            * conj(
                                (
                                    self.V_r(n, r, ac)
                                    - n**2 * self.V_theta(n, r, ac)
                                )
                                / r
                                - self.k_f**2 * self.phi(n, r, ac),
                            )
                            + self.V_theta(m, r, ac)
                            * conj(
                                (
                                    self.d_V_r(n, r, ac)
                                    - n**2 * self.d_V_theta(n, r, ac)
                                )
                                / r
                                - (
                                    self.V_r(n, r, ac)
                                    - n**2 * self.V_theta(n, r, ac)
                                )
                                / r**2
                                - self.k_f**2 * self.d_phi(n, r, ac),
                            )
                        )
                    )

        second_sum = complex(0)
        for n in self.range_2_N_max:
            for m in full_range(1, int(n / 2)):
                second_inner_sum = complex(0)
                for k in full_range(abs(n - 2 * m - l), n - 2 * m + l):
                    cg_coef_1 = cg(n - 2 * m, 0, l, 0, k, 0)
                    cg_coef_2 = cg(n - 2 * m, 0, l, 1, k, 1)
                    if k >= 1 and cg_coef_1 and cg_coef_2:
                        second_inner_sum += (
                            sqrt(k * (k + 1))
                            * cg_coef_1
                            * cg_coef_2
                            / (2 * k + 1)
                            * (
                                (
                                    conj(self.d_V_theta(n, r, ac))
                                    * self.V_theta(k, r, ac)
                                    + conj(self.V_theta(n, r, ac))
                                    * self.d_V_theta(k, r, ac)
                                )
                                / r
                                - (
                                    conj(self.V_theta(n, r, ac))
                                    * self.V_theta(k, r, ac)
                                )
                                / r**2
                            )
                        )
                second_sum += (2 * n - 4 * m + 1) * second_inner_sum

        return (
            (2 * l + 1)
            / (2 * sqrt(l * (l + 1)))
            * (first_sum + second_sum).real
        )

    def E(self, l: int, r: float, ac: bool) -> float:
        return (
            -self.rho_f
            / (self.eta_f * r)
            * (
                self.gamma(l, r, ac)
                + r * self.d_gamma(l, r, ac)
                - self.beta(l, r, ac)
            )
        )

    def F(self, n: int, m: int, r: float, ac: bool) -> complex:
        return (
            1j
            * conj(self.k_f**2)
            / r
            * (
                conj(2 * self.phi(n, r, ac) + r * self.d_phi(n, r, ac))
                * self.V_r(m, r, ac)
                + r * conj(self.phi(n, r, ac)) * self.d_V_r(m, r, ac)
            )
        )

    def G(self, n: int, m: int, r: float, ac: bool) -> complex:
        return (
            1j
            * conj(self.k_f**2)
            / r
            * conj(self.phi(n, r, ac))
            * self.V_theta(m, r, ac)
        )

    # -------------------------------------------------------------------------
    # Integrals
    # -------------------------------------------------------------------------

    def compute_C_0(self) -> NDArray:

        l_range = self.range_1_N_max

        C_0 = np.zeros((6 + 1, self.cls.N_max + 1))

        # parameters
        inf = self.cls.inf
        int_eps = self.cls.integration_rel_eps

        for l in l_range:

            # C_1l0
            C_0[1, l] = 0

            # C_2l0
            kernel_C_2l0 = lambda s: s ** (1 - l) * (
                self.alpha(l, s, ac=False) - self.alpha(l, s, ac=True)
            )
            integral_C_2l0, error_C_2l0 = integrate(
                kernel_C_2l0,
                self.R_0,
                inf,
                rel_eps=int_eps,
            )
            C_0[2, l] = -1 / (2 * l + 1) * integral_C_2l0

            # C_5l0
            kernel_C_5l0 = lambda s: s ** (3 - l) * (
                self.E(l, s, ac=False) - self.E(l, s, ac=True)
            )
            integral_C_5l0, error_C_5l0 = integrate(
                kernel_C_5l0,
                self.R_0,
                inf,
                rel_eps=int_eps,
            )
            C_0[5, l] = 1 / (2 * (2 * l - 1) * (2 * l + 1)) * integral_C_5l0

            # C_6l0
            kernel_C_6l0 = lambda s: s ** (1 - l) * (
                self.E(l, s, ac=False) - self.E(l, s, ac=True)
            )
            integral_C_6l0, error_C_6l0 = integrate(
                kernel_C_6l0,
                self.R_0,
                inf,
                rel_eps=int_eps,
            )
            C_0[6, l] = -1 / (2 * (2 * l + 1) * (2 * l + 3)) * integral_C_6l0

            # C_3l0
            C_0[3, l] = 0.5 * (
                self.R_0**l
                * (
                    self.V_S_r(l, self.R_0, ac=False) / (l + 1)
                    + self.V_S_theta(l, self.R_0, ac=False)
                )
                + (2 * l + 1)
                * self.R_0 ** (2 * l - 1)
                * (C_0[2, l] / (l + 1) - C_0[5, l])
                - (2 * l + 3) * self.R_0 ** (2 * l + 1) * C_0[6, l]
            )

            # C_4l0
            C_0[4, l] = 0.5 * (
                self.R_0 ** (l + 2)
                * (
                    (2 - l) * self.V_S_r(l, self.R_0, ac=False) / (l * (l + 1))
                    - self.V_S_theta(l, self.R_0, ac=False)
                )
                - (2 * l - 1)
                * self.R_0 ** (2 * l + 1)
                * (C_0[2, l] / (l + 1) - C_0[5, l])
                + (2 * l + 1) * self.R_0 ** (2 * l + 3) * C_0[6, l]
            )

        return C_0

    # -------------------------------------------------------------------------
    # Stokes drift velocity
    # -------------------------------------------------------------------------

    def V_S_r(self, l: int, r: float, ac: bool) -> float:
        first_sum = complex(0)
        for n in self.range_N_max:
            first_inner_sum = complex(0)
            for m in full_range(abs(l - n), l + n):
                if cg_coef := cg(l, 0, n, 0, m, 0):
                    first_inner_sum += (
                        cg_coef**2 / (2 * m + 1) * conj(self.d_V_r(m, r, ac))
                    )
            first_sum += self.V_r(n, r, ac) * first_inner_sum
        second_sum = complex(0)
        for n in self.range_1_N_max:
            for m in full_range(1, int((n + 1) / 2)):
                second_inner_sum = complex(0)
                for k in full_range(
                    abs(l - (n - 2 * m + 1)),
                    l + n - 2 * m + 1,
                ):
                    cg_coef = cg(l, 0, n - 2 * m + 1, 0, k, 0)
                    if cg_coef:
                        factor = cg_coef**2 / (2 * k + 1)
                        term_1 = (
                            (k + 1)
                            * (k + 2)
                            / (2 * k + 3)
                            * conj(self.V_r(k + 1, r, ac))
                        )
                        if k <= 1:
                            second_inner_sum += factor * term_1
                            continue
                        term_2 = (
                            k
                            * (k - 1)
                            / (2 * k - 1)
                            * conj(self.V_r(k - 1, r, ac))
                        )
                        second_inner_sum += factor * (term_1 - term_2)
                second_sum += (
                    (2 * n - 4 * m + 3)
                    * self.V_theta(n, r, ac)
                    / r
                    * second_inner_sum
                )
        return (
            (2 * l + 1)
            / (2 * self.omega)
            * (1j * (first_sum + second_sum)).real
        )

    def V_S_theta(self, l: int, r: float, ac: bool) -> float:
        if l == 0:
            return 0

        first_sum = complex(0)
        for n in self.range_N_max:
            for m in full_range(abs(n - l), n + l):
                cg_coef_1 = cg(n, 0, l, 0, m, 0)
                cg_coef_2 = cg(n, 0, l, 1, m, 1)
                if m >= 1 and cg_coef_1 and cg_coef_2:
                    first_sum += (
                        sqrt(m * (m + 1))
                        * cg_coef_1
                        * cg_coef_2
                        / (2 * m + 1)
                        * (
                            self.V_r(n, r, ac) * conj(self.d_V_theta(m, r, ac))
                            + self.V_theta(m, r, ac)
                            * conj(
                                self.V_r(n, r, ac)
                                - n**2 * self.V_theta(n, r, ac),
                            )
                            / r
                        )
                    )
        second_sum = self._V_S_theta_second_sum(l, r, ac)
        return (
            (2 * l + 1)
            / (2 * self.omega * sqrt(l * (l + 1)))
            * (1j * (first_sum + second_sum))
        ).real

    def _V_S_theta_second_sum(self, l: int, r: float, ac: bool) -> complex:

        second_sum = complex(0)
        for n in self.range_2_N_max:
            second_middle_sum = complex(0)
            for m in full_range(1, int(n / 2)):
                second_inner_sum = complex(0)
                for k in full_range(abs(n - 2 * m - l), n - 2 * m + l):
                    cg_coef_1 = cg(n - 2 * m, 0, l, 0, k, 0)
                    cg_coef_2 = cg(n - 2 * m, 0, l, 1, k, 1)
                    if k >= 1 and cg_coef_1 and cg_coef_2:
                        second_inner_sum += (
                            sqrt(k * (k + 1))
                            * cg_coef_1
                            * cg_coef_2
                            * self.V_theta(k, r, ac)
                            / (2 * k + 1)
                        )
                second_middle_sum += (2 * n - 4 * m + 1) * second_inner_sum
            second_sum += conj(self.V_theta(n, r, ac)) / r * second_middle_sum

        return second_sum

    # -------------------------------------------------------------------------
    # Tests
    # -------------------------------------------------------------------------

    def test_infinity_not_set(self):
        self.do_testing(self.inf, lambda: self.cls.inf)

    def test_infinity_set(self):
        # inf type
        self.cls.inf_type = "radius"
        self.do_testing(
            lambda: self.cls.inf_factor * self.R_0,
            lambda: self.cls.inf,
        )
        self.cls.inf_type = "delta"
        self.cls.inf_factor = 80
        self.do_testing(
            lambda: self.cls.inf_factor * self.delta,
            lambda: self.cls.inf,
        )
        self.cls.inf_type = 1
        self.do_testing(
            lambda: self.cls.inf_factor,
            lambda: self.cls.inf,
        )

    def test_wrong_inf_type(self):
        self.cls.inf_type = "blabla"
        self.assertRaises(ValueError, lambda: self.cls.inf)

    def test_F_G_dependent(self) -> None:
        self.parameters.list_parameters.remove(self.parameters._R_0)

        for m in range(1, self.runs + 1):
            factor = self.parameters.rng.uniform(2, 10)
            dict_methods = dict()
            dict_methods["F"] = (m, factor * self.R_0, True)
            dict_methods["G"] = (m, factor * self.R_0, True)
            self._test_methods_n(dict_methods)

    def test_alpha_beta_gamma(self) -> None:
        radii = self.R_0 * np.linspace(1, 5, self.n_runs)
        for r in radii:
            for n in range(1, 5):
                self.assertAlmostEqual(
                    self.alpha(n, r, False),
                    self.cls.alpha(n, r, False),
                    threshold=1e-8,
                )
                self.assertAlmostEqual(
                    self.beta(n, r, False),
                    self.cls.beta(n, r, False),
                    1e-6,
                )
                self.assertAlmostEqual(
                    self.gamma(n, r, False),
                    self.cls.gamma(n, r, False),
                    threshold=1e-6,
                )
                self.assertAlmostEqual(
                    self.d_gamma(n, r, False),
                    self.cls.d_gamma(n, r, False),
                    threshold=1e-6,
                )

    def test_gamma_error(self):
        self.assertRaises(ValueError, self.cls.gamma, 0, 1, True)
        self.assertRaises(ValueError, self.cls.d_gamma, 0, 1, True)

    def test_E(self) -> None:
        radii = self.R_0 * np.linspace(1, 5, self.n_runs)
        for r in radii:
            for n in range(1, 5):
                self.assertAlmostEqual(
                    self.E(n, r, False),
                    self.cls.E(n, r, False),
                    threshold=1e-6,
                )

    def test_integrals(self):
        self.cls.N_max = 1
        C_0 = self.compute_C_0()
        for n in range(1, 7):
            for m in self.range_1_N_max:
                with self.subTest(msg=f"C_{n}{m}0"):
                    self.assertAlmostEqual(self.cls.C_0[n, m], C_0[n, m])

    def test_rel_eps(self):
        self.cls.N_max = 1
        self.cls.integration_rel_eps = 1e-3
        C_first = self.cls.C_0[3, 1]
        self.cls.integration_rel_eps = 1e-9
        C_second = self.cls.C_0[3, 1]
        self.assertAlmostEqual(C_first, C_second)

    def test_stokes_drift_velocity(self) -> None:
        radii = self.R_0 * np.linspace(1, 5, self.n_runs)
        for r in radii:
            for n in range(0, 5):
                first = self._V_S_theta_second_sum(n, r, False)
                second = self.cls._V_S_theta_second_sum(n, r, False)
                if abs(first) > 1e-20 and abs(second) > 1e-20:
                    self.assertAlmostEqual(first, second, threshold=1e-8)
                first = self.V_S_theta(n, r, False)
                second = self.cls.V_S_theta(n, r, False)
                if abs(first) > 1e-20 and abs(second) > 1e-20:
                    self.assertAlmostEqual(first, second, threshold=1e-8)
                first = self.V_S_r(n, r, False)
                second = self.cls.V_S_r(n, r, False)
                if abs(first) > 1e-20 and abs(second) > 1e-20:
                    self.assertAlmostEqual(first, second, threshold=1e-8)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from abc import ABC, abstractmethod

import numpy as np

from osaft.core.variable import ActiveVariable, PassiveVariable
from osaft.plotting.grid import (
    CartesianFluidGrid,
    CartesianGrid,
    CartesianScattererGrid,
    MaskedCartesianGrid,
    PolarGrid,
)
from osaft.solutions.base_scattering import BaseScattering
from osaft.solutions.base_streaming import BaseStreaming

NDArray = np.ndarray


class VelocityFieldData(ABC):
    """Container for plotting data for scattering and streaming plots

    :param sol: solution
    :param r_min: start point for radial plotting range in [m]
    :param r_max: end point for tangential plotting range  in [m]
    :param res: resolution, if tuple (radial resolution, tangential resolution)
    :param n_quiver_points: anchor points along z for quiver arrows
    :param symmetric: if ``True`` the dataset is assumed to be symmetric
    """

    _EPS_r_min = 1e-30

    def __init__(
        self,
        sol: BaseScattering | BaseStreaming,
        r_min: float,
        r_max: float,
        res: int | tuple[int, int],
        n_quiver_points: int | tuple[int, int],
        symmetric,
    ) -> None:
        """Constructor method"""

        # Parameters for the coordinate grid
        self._r_min = PassiveVariable(r_min, "r min")
        self._r_max = PassiveVariable(r_max, "r max")
        self._resolution = PassiveVariable(res, "plotting resolution")
        self._n_quiver_points = PassiveVariable(
            n_quiver_points,
            "quiver resolution along z",
        )
        self._symmetric = PassiveVariable(symmetric)

        # Solution to be plotted
        self._sol = PassiveVariable(sol, "solution for plotting")
        self._mode = PassiveVariable(None, "mode")

        # Dependent Variables: Data containers for solution
        self._x = ActiveVariable(self._reset_real_array, "x coordinate")
        self._z = ActiveVariable(self._reset_real_array, "z coordinate")
        self._u = ActiveVariable(
            self._reset_complex_array,
            "velocity/displacement in x-direction",
        )
        self._w = ActiveVariable(
            self._reset_complex_array,
            "velocity/displacement in y-direction",
        )
        self._v_norm = ActiveVariable(
            self._reset_real_array,
            "norm of the velocity / displacement",
        )
        # Dependent Variables: Data containers for quiver solution
        self._x_quiver = ActiveVariable(
            self._reset_real_array_quiver,
            "x coordinate of quiver arrows",
        )
        self._z_quiver = ActiveVariable(
            self._reset_real_array_quiver,
            "z coordinate of quiver arrows",
        )
        self._u_quiver = ActiveVariable(
            self._reset_complex_array_quiver,
            "velocity in z-direction (quiver)",
        )
        self._w_quiver = ActiveVariable(
            self._reset_complex_array_quiver,
            "velocity in x-direction (quiver)",
        )

        # Dependent Variables: Coordinate grid
        self._grid = ActiveVariable(self._get_grid)
        self._grid_quiver = ActiveVariable(self._get_grid_quiver)

        # Setting links
        self._grid.is_computed_by(
            self._r_min,
            self._r_max,
            self._resolution,
            self._symmetric,
        )
        self._grid_quiver.is_computed_by(
            self._r_min,
            self._r_max,
            self._n_quiver_points,
            self._symmetric,
        )
        # contour solution
        self._x.is_computed_by(self._grid)
        self._z.is_computed_by(self._grid)
        self._u.is_computed_by(self._sol, self._grid, self._mode)
        self._w.is_computed_by(self._sol, self._grid, self._mode)
        self._v_norm.is_computed_by(self._sol, self._grid, self._mode)
        # quiver solution
        self._x_quiver.is_computed_by(self._grid_quiver)
        self._z_quiver.is_computed_by(self._grid_quiver)
        self._u_quiver.is_computed_by(self._sol, self._grid_quiver, self._mode)
        self._w_quiver.is_computed_by(self._sol, self._grid_quiver, self._mode)

    # -------------------------------------------------------------------------
    # API
    # -------------------------------------------------------------------------

    def get_velocity_magnitude(self) -> tuple[NDArray, NDArray, NDArray]:
        """Returns the velocity magnitudes

        returns the coordinates of the points of the plotting grid `x`,
        `y` and the magnitude of the velocity field at
        these points
        """
        return self.x, self.z, self.u_norm  # pragma: no cover

    def get_velocity_vector(self) -> tuple[NDArray, NDArray, NDArray, NDArray]:
        """Returns the velocity vectors

        returns the coordinates of the points of the plotting grid `x`,
        `y` and the velocity field vectors at
        these points
        """
        return self.x, self.z, self.u, self.w  # pragma: no cover

    def get_velocity_vector_quiver(
        self,
    ) -> tuple[NDArray, NDArray, NDArray, NDArray]:
        """Returns the velocity vectors and coordinates for the quiver plot"""
        return (
            self.z_quiver,
            self.x_quiver,  # pragma: no cover
            self.u_quiver,
            self.w_quiver,  # pragma: no cover
        )

    # -------------------------------------------------------------------------
    # Getters and Setters for Independent Variables
    # -------------------------------------------------------------------------

    @property
    def sol(self) -> BaseScattering | BaseStreaming:
        """Solution for plotting

        :getter: returns the end point for radial plotting range
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._sol.value

    @sol.setter
    def sol(self, value: BaseScattering | BaseStreaming) -> None:
        self._sol.value = value

    @property
    def r_min(self) -> float:
        """Endpoint for radial plotting range [m]

        :getter: returns the end point for radial plotting range
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._r_min.value

    @r_min.setter
    def r_min(self, value: float) -> None:
        self._r_min.value = value

    @property
    def r_max(self) -> float:
        """Start point for radial plotting range [m]

        :getter: returns the start point for radial plotting range
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._r_max.value

    @r_max.setter
    def r_max(self, value: float) -> None:
        self._r_max.value = value

    @property
    def n_quiver_points(self) -> int:
        """Number of anchor points for quiver arrows

        :getter: returns n_point
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._n_quiver_points.value

    @n_quiver_points.setter
    def n_quiver_points(self, value: int) -> None:
        self._n_quiver_points.value = value

    @property
    def resolution(self) -> int | tuple[int, int]:
        """Resolution for scattering / streaming plots

        `resolution` can be an `int` when the resolution in radial and
        tangential direction is the same, or a `tuple` of two `int`
        for independent resolutions (radial resolution, tangential_resolution)
        :getter: returns plotting resolution
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._resolution.value

    @resolution.setter
    def resolution(self, value: int | tuple[int, int]) -> None:
        self._resolution.value = value

    @property
    def mode(self) -> int:
        """Mode that is plotted. If `None` all modes up to `sol.N_max`
        are superimposed.
        :getter: returns the mode
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._mode.value

    @property
    def symmetric(self) -> bool:
        """Symmetry option of the grid

        If ``True`` the velocity field is assumed to be symmetric.
        :getter: returns plotting resolution
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._symmetric.value

    @symmetric.setter
    def symmetric(self, value: bool) -> None:
        self._symmetric = value

    # -------------------------------------------------------------------------
    # Getters for Dependent Variables (contour solution)
    # -------------------------------------------------------------------------

    @property
    def grid(self) -> PolarGrid:
        """Plotting grid"""
        return self._grid.value

    @property
    def x(self) -> NDArray:
        """x-coordinates of the plotting grid in [m]"""
        if np.all(self._x.value == 0):
            self._compute_data()
        return self._x.value

    @property
    def z(self) -> NDArray:
        """z-coordinates of the plotting grid in [m]"""
        if np.all(self._z.value == 0):
            self._compute_data()  # pragma: no cover
        return self._z.value

    @property
    def u(self) -> NDArray:
        """Velocity / displacement in x-direction in [m/s]"""
        if np.all(self._u.value == 0):
            self._compute_data()
        return self._u.value

    @property
    def w(self) -> NDArray:
        """Velocity / displacement in z-direction in [m/s]"""
        if np.all(self._w.value == 0):
            self._compute_data()  # pragma: no cover
        return self._w.value

    @property
    def u_norm(self) -> NDArray:
        """Norm of the velocity / displacement in z-direction in [m/s]

        norm = abs(u) ** 2 + abs(w) ** 2
        """
        if np.all(self._v_norm.value == 0):
            self._compute_data()
        return self._v_norm.value

    # -------------------------------------------------------------------------
    # Getters for Dependent Variables (quiver solution)
    # -------------------------------------------------------------------------

    @property
    def grid_quiver(self) -> CartesianGrid:
        """Plotting grid"""
        return self._grid_quiver.value

    @property
    def x_quiver(self) -> NDArray:
        """x-coordinates of the quiver plotting grid in [m]"""
        if np.all(self._x_quiver.value == 0):
            self._compute_data_quiver()
        return self._x_quiver.value

    @property
    def z_quiver(self) -> NDArray:
        """z-coordinates of the quiver plotting grid in [m]"""
        if np.all(self._z_quiver.value == 0):
            self._compute_data_quiver()  # pragma: no cover
        return self._z_quiver.value

    @property
    def u_quiver(self) -> NDArray:
        """Velocity / displacement in x-direction in [m/s]"""
        if np.all(self._u_quiver.value == 0):
            self._compute_data_quiver()
        return self._u_quiver.value

    @property
    def w_quiver(self) -> NDArray:
        """Velocity / displacement in z-direction in [m/s]"""
        if np.all(self._w_quiver.value == 0):
            self._compute_data_quiver()  # pragma: no cover
        return self._w_quiver.value

    # -------------------------------------------------------------------------
    # Methods for dependent variables
    # -------------------------------------------------------------------------

    def _reset_real_array(self) -> NDArray:
        """Resets real arrays for coordinates and norm"""
        return np.zeros(
            self.grid.res_r * self.grid.res_theta,
            dtype=np.float64,
        )

    def _reset_real_array_quiver(self) -> NDArray:
        """Resets real arrays (quiver) for coordinates and norm"""
        return np.zeros(
            len(self.grid_quiver.flat_mesh_x),
        )

    def _reset_complex_array(self) -> NDArray:
        """Resets complex arrays for complex velocities and norm"""
        return np.zeros(
            self.grid.res_r * self.grid.res_theta,
            dtype=np.complex128,
        )

    def _reset_complex_array_quiver(self) -> NDArray:
        """Resets complex arrays (quiver) for complex velocities and norm"""
        return np.zeros(
            len(self.grid_quiver.flat_mesh_x),
            dtype=np.complex128,
        )

    # -------------------------------------------------------------------------
    # Private Methods
    # -------------------------------------------------------------------------

    def _get_grid(self) -> PolarGrid:
        """Returns a new coordinate grid"""
        return PolarGrid(
            self.r_min,
            self.r_max,
            self.resolution,
            self.symmetric,
        )

    def _get_grid_quiver(self) -> MaskedCartesianGrid:
        """Returns a new coordinate grid for the quiver plots"""
        if self.r_min <= self._EPS_r_min:
            return CartesianScattererGrid(
                self.r_max,
                self.n_quiver_points,
                self.symmetric,
            )
        else:
            return CartesianFluidGrid(
                self.sol.R_0,
                self.r_max,
                self.n_quiver_points,
                self.symmetric,
            )

    @abstractmethod
    def _compute_data(self) -> tuple[NDArray, NDArray]:
        """Compute the velocity data and store it in the attributes"""
        pass  # pragma: no cover

    @abstractmethod
    def _compute_data_quiver(self) -> tuple[NDArray, NDArray]:
        """Compute the velocity data and store it in the respective
        attributes"""
        pass  # pragma: no cover


if __name__ == "__main__":
    pass



import unittest

from osaft.core.functions import pi
from osaft.plotting.datacontainers.scattering_datacontainer import (
    FluidScatteringData,
)
from osaft.tests.solution_factory import SolutionFactory


class TestBaseScatteringDatacontainer(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()

        self.sol = SolutionFactory().yosioka_1955_scattering()

        self.cls = FluidScatteringData(
            self.sol,
            5 * self.sol.R_0,
        )

    def test_setters_and_getters(self):
        # Change solution
        sol_2 = SolutionFactory().doinikov_1994_rigid_scattering()

        self.cls.sol = sol_2

        # Change r_min
        self.cls.r_min = 2e-6
        self.assertEqual(self.cls.r_min, 2e-6)

        # Change r_max
        self.cls.r_max = 10e-6
        self.assertEqual(self.cls.r_max, 10e-6)

        # Change theta min
        self.cls.theta_min = pi / 4
        self.assertEqual(self.cls.theta_min, pi / 4)

        # Change theta max
        self.cls.theta_max = pi / 2
        self.assertEqual(self.cls.theta_max, pi / 2)

        # Change n_quiver_points
        self.cls.n_quiver_points = 42
        self.assertEqual(self.cls.n_quiver_points, 42)

    def test_change_resolution(self):
        self.cls.resolution = 200
        self.assertEqual(self.cls.grid.res_r, 200)


if __name__ == "__main__":
    unittest.main()



import unittest

from osaft import WaveType
from osaft.core.backgroundfields import WrongWaveTypeError
from osaft.core.functions import BesselFunctions as Bf
from osaft.core.functions import conj, pi, sin
from osaft.tests.basetest_arf import HelperCompareARF
from osaft.tests.basetest_numeric import BaseTestSolutions
from osaft.tests.solution_factory import SolutionFactory


class TestARF(BaseTestSolutions):
    def setUp(self) -> None:
        super().setUp()

        self.cls = SolutionFactory().yosioka_1955_arf()

        self.list_cls = [self.cls]

    # -------------------------------------------------------------------------
    # Methods
    # -------------------------------------------------------------------------

    @property
    def F(self):
        sigma = self.cls.sigma
        lamb = self.cls.lambda_rho
        x_s = self.cls.x_s

        if self.wave_type == WaveType.STANDING:
            if self.cls.bubble_solution:
                out = sigma * x_s**3 * (3 * lamb - x_s**2)
                out /= sigma**2 * x_s**6 + (3 * lamb - x_s**2) ** 2
                return out
            else:
                out = lamb + (2 * (lamb - 1) / 3)
                out /= 1 + 2 * lamb
                out -= 1 / (3 * lamb * sigma**2)
                return out
        else:
            if self.cls.bubble_solution:
                pass
            else:
                out = (lamb - (1 + 2 * lamb) / (3 * lamb * sigma**2)) ** 2
                out += 2 / 9 * (1 - lamb) ** 2
                out *= 1 / (1 + 2 * lamb) ** 2
                return out

    def small_particle_ARF(self):
        if self.wave_type == WaveType.STANDING:
            if self.cls.bubble_solution:
                return self._small_bubble_standing()
            else:
                return self._small_particle_standing()
        else:
            if self.cls.bubble_solution:
                return self._small_bubble_travelling()
            else:
                return self._small_particle_travelling()

    def _small_particle_standing(self):
        k = self.cls.k_f
        E = self.cls.E_ac
        R = self.R_0
        F = self.F
        h = self.cls.position
        return pi * R**3 * 4 * k * E * sin(2 * h) * F

    def _small_particle_travelling(self):
        R = self.R_0
        I = self.cls.I_ac
        c = self.c_f
        F = self.F
        x = self.cls.x_f
        return pi * R**2 * 4 * x**4 * I / c * F

    def _small_bubble_travelling(self):
        I = self.cls.I_ac
        sigma = self.cls.sigma
        lamb = self.cls.lambda_rho
        x_s = self.cls.x_s
        R = self.R_0
        c = self.c_f
        res = 3 * lamb - x_s**2
        out = 4 * pi * R**2 * x_s**4 * I
        out /= c * (sigma**2 * x_s**6 + res**2)
        return out

    def _small_bubble_standing(self):
        E = self.cls.E_ac
        k = self.cls.k_f
        h = self.position
        F = self.F
        return -4 * pi / k**2 * E * sin(2 * h) * F

    def M_n(self, n):
        B_n = self.cls.B_n(n)
        x_s = self.cls.x_s
        return (-1) ** n / (2 * n + 1) * B_n * Bf.besselj(n, x_s)

    def K_n(self, n):
        B_n = self.cls.B_n(n)
        k_s = self.cls.k_s
        x_s = self.cls.x_s
        return (-1) ** n / (2 * n + 1) * B_n * k_s * Bf.d1_besselj(n, x_s)

    def compute_arf(self):
        a = self.R_0
        rho = self.rho_f
        lamb = self.cls.lambda_rho
        k = self.cls.k_f
        s = 0
        for n in range(self.cls.N_max):
            s1 = (
                -2
                * pi
                * a**2
                * rho
                * 2
                * (n + 1)
                * (self.K_n(n) * conj(self.K_n(n + 1))).real
                / 2
            )
            s2 = (
                2
                * pi
                * lamb**2
                * rho
                * 2
                * n
                * (n + 1)
                * (n + 2)
                * (self.M_n(n) * conj(self.M_n(n + 1))).real
                / 2
            )
            s3 = (
                -2
                * pi
                * a
                * lamb
                * rho
                * 2
                * (n + 1)
                * (n + 2)
                * (self.K_n(n) * conj(self.M_n(n + 1))).real
                / 2
                + 2
                * pi
                * a
                * lamb
                * rho
                * 2
                * n
                * (n + 1)
                * (self.M_n(n) * conj(self.K_n(n + 1))).real
                / 2
            )
            s4 = (
                -2
                * pi
                * k**2
                * a**2
                * lamb**2
                * rho
                * 2
                * (n + 1)
                * (self.M_n(n) * conj(self.M_n(n + 1))).real
                / 2
            )
            s += s1 + s2 + s3 + s4
        if self.cls.wave_type == WaveType.STANDING:
            return -s
        return s

    # -------------------------------------------------------------------------
    # Test
    # -------------------------------------------------------------------------

    def test_F(self) -> None:
        try:
            self.do_testing(
                func_1=lambda: self.F,
                func_2=lambda: self.cls.F,
            )
        except ValueError:
            pass

    def test_small_particle_solution(self):
        self.cls.small_particle = True
        self.do_testing(
            self.cls.compute_arf,
            self.small_particle_ARF,
        )

    def test_small_bubble_solution(self):
        self.cls.small_particle = True
        self.cls.bubble_solution = True
        self.do_testing(
            self.cls.compute_arf,
            self.small_particle_ARF,
        )

    def test_value_error_wrong_wavetype(self):
        self.cls.wave_type = "wrong_wavetype"

        # Small Particle
        self.cls.small_particle = True
        self.assertRaises(
            WrongWaveTypeError,
            self.cls.compute_arf,
        )

        # Small Bubble
        self.cls.bubble_solution = True
        self.assertRaises(
            WrongWaveTypeError,
            self.cls.compute_arf,
        )

    def test_large_bubble_error(self):
        self.cls.bubble_solution = True
        self.cls.small_particle = False
        self.assertRaises(
            ValueError,
            self.cls.compute_arf,
        )

    def test_value_error_F(self):
        # Small Particle
        self.cls.bubble_solution = True
        self.cls.wave_type = WaveType.TRAVELLING
        self.assertRaises(WrongWaveTypeError, lambda: self.cls.F)

    def test_M_n(self):
        for n in range(self.n_runs):
            self.do_testing(self.cls.M_n, self.M_n, n, n)

    def test_K_n(self):
        for n in range(self.n_runs):
            self.do_testing(self.cls.K_n, self.K_n, n, n)

    def test_general_arf(self):
        self.do_testing(
            self.compute_arf,
            self.cls.compute_arf,
        )

    def test_run_large_particle_warning(self):
        self.cls.small_particle = True
        self.parameters.R_0 = 2 * pi * self.cls.fluid.lambda_f
        self.assign_parameters()
        self.cls.compute_arf()
        self.cls.bubble_solution = True
        self.assign_parameters()
        self.cls.compute_arf()


class TestCompareToKing(BaseTestSolutions, HelperCompareARF):
    def setUp(self) -> None:

        BaseTestSolutions.setUp(self)

        # Setting speed of sound in particle to be large
        self.parameters.c_s = 1e3 * self.c_f
        self.parameters._c_s.low = 1e3 * self.c_f

        self.arf_compare_threshold = 1e-4

        self.cls = SolutionFactory().yosioka_1955_arf()
        self.compare_cls = SolutionFactory().king_1934_arf()
        self.list_cls = [self.cls, self.compare_cls]
        self.assign_parameters()


class TestCompareToGorkov(BaseTestSolutions, HelperCompareARF):
    def setUp(self) -> None:

        BaseTestSolutions.setUp(self)

        self.arf_compare_threshold = 1e-3

        self.cls = SolutionFactory().yosioka_1955_arf()
        self.compare_cls = SolutionFactory().gorkov_1962_arf()

        # Setting size of the particle to be small
        self.parameters.R_0 = self.cls.fluid.lambda_f * 1e-4
        self.parameters._R_0.high = self.cls.fluid.lambda_f * 1e-3
        self.list_cls = [self.cls, self.compare_cls]
        self.assign_parameters()


class TestCompareToSmallParticle(BaseTestSolutions, HelperCompareARF):
    def setUp(self) -> None:

        BaseTestSolutions.setUp(self)

        self.arf_compare_threshold = 5e-3

        self.cls = SolutionFactory().yosioka_1955_arf()
        self.compare_cls = SolutionFactory().yosioka_1955_arf()
        self.compare_cls.small_particle = True

        # Setting size of the particle to be small
        self.parameters.R_0 = self.cls.fluid.lambda_f * 1e-4
        self.parameters._R_0.high = self.cls.fluid.lambda_f * 1e-3
        self.list_cls = [self.cls, self.compare_cls]
        self.assign_parameters()


class TestCompareSmallBubble(BaseTestSolutions, HelperCompareARF):
    def setUp(self) -> None:
        BaseTestSolutions.setUp(self)

        self.arf_compare_threshold = 5e-3

        self.cls = SolutionFactory().yosioka_1955_arf()
        self.compare_cls = SolutionFactory().yosioka_1955_arf()
        self.compare_cls.small_particle = True
        self.compare_cls.bubble_solution = True

        # Setting size of the particle to be small
        self.parameters.R_0 = self.cls.fluid.lambda_f * 1e-4
        self.parameters._R_0.high = self.cls.fluid.lambda_f * 1e-3

        # Set density to be small inside to be small
        self.parameters.rho_s = self.cls.x_f**2 * self.rho_f
        self.parameters._rho_s.high = 1.1 * self.rho_s

        self.list_cls = [self.cls, self.compare_cls]


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
from matplotlib import pyplot as plt

from osaft.plotting.grid import BaseGrid, PolarGrid


def plot_grid(grid: BaseGrid) -> None:
    fig, ax = plt.subplots()
    ax.scatter(grid.mesh_z, grid.mesh_x)
    ax.set_aspect("equal")
    fig, ax = plt.subplots(subplot_kw={"projection": "polar"})
    ax.scatter(grid.mesh_theta, grid.mesh_r)


class TestBaseGrid(unittest.TestCase):
    """Testing BaseGrid methods on PolarGrid, since BaseGrid is an ABC"""

    # -------------------------------------------------------------------------
    # Resolution unpacking
    # -------------------------------------------------------------------------

    def test_int_resolution(self) -> None:

        self.grid = PolarGrid(0, 1, 10, False)
        res_x1 = self.grid._res_x1
        res_x2 = self.grid._res_x2
        self.assertEqual(res_x1, 10)
        self.assertEqual(res_x2, 10)

    def test_tuple_resolution(self) -> None:

        self.grid = PolarGrid(0, 1, (10, 20), False)
        res_x1 = self.grid._res_x1
        res_x2 = self.grid._res_x2
        self.assertEqual(res_x1, 10)
        self.assertEqual(res_x2, 20)

    def test_even_int_resolution_upper(self) -> None:

        self.grid = PolarGrid(0, 1, 10, True)
        res_x1 = self.grid._res_x1
        res_x2 = self.grid._res_x2
        self.assertEqual(res_x1, 10)
        self.assertEqual(res_x2, 5)

    def test_odd_int_resolution_upper(self) -> None:

        self.grid = PolarGrid(0, 1, 11, True)
        res_x1 = self.grid._res_x1
        res_x2 = self.grid._res_x2
        self.assertEqual(res_x1, 11)
        self.assertEqual(res_x2, 5)

    def test_ValuerError(self) -> None:
        self.assertRaises(
            ValueError,
            PolarGrid,
            r_min=0,
            r_max=1,
            resolution=3 * [10],
            is_upper=True,
        )

    # -------------------------------------------------------------------------
    # Flatten mesh
    # -------------------------------------------------------------------------

    def test_flatten_mesh(self):

        self.grid = PolarGrid(0, 1, 11, True)
        array = np.random.random((3, 3))
        flat_array = array.flatten()

        self.grid._mesh_r = array
        self.grid._mesh_theta = array
        self.grid._mesh_x = array
        self.grid._mesh_z = array

        np.testing.assert_array_equal(self.grid.flat_mesh_r, flat_array)
        np.testing.assert_array_equal(self.grid.flat_mesh_theta, flat_array)
        np.testing.assert_array_equal(self.grid.flat_mesh_x, flat_array)
        np.testing.assert_array_equal(self.grid.flat_mesh_z, flat_array)


if __name__ == "__main__":
    unittest.main()



import unittest

from osaft.core.functions import BesselFunctions as Bf
from osaft.core.functions import cos, exp, sin
from osaft.tests.basetest_scattering import (
    HelperCompareScattering,
    HelperScattering,
)
from osaft.tests.solution_factory import SolutionFactory
from osaft.tests.solutions.basedoinikov1994.test_base import (
    BaseTestDoinikov1994,
    TestBaseDoinikov1994,
)


class TestScattering(BaseTestDoinikov1994, HelperScattering):
    def setUp(self) -> None:
        super().setUp()

        self._v_particle_threshold = 1e-5
        self._v_fluid_threshold = 1e-3
        self._v_boundary_conditions = 1e-6

        self.cls = SolutionFactory().doinikov_1994_rigid_scattering()

    def V_r_sc(self, n, r) -> complex:
        # grad(phi)
        out = self.cls.k_f * self.alpha_n(n)
        out *= Bf.d1_hankelh1(n, self.cls.k_f * r)
        # curl(Psi)
        tmp = n * (n + 1) / r * self.beta_n(n)
        tmp *= Bf.hankelh1(n, self.cls.k_v * r)

        out -= tmp

        out *= self.cls.A_in(n)
        return out

    def V_theta_sc(self, n, r) -> complex:
        # curl(Psi)
        arg_v = self.cls.k_v * r
        out = Bf.hankelh1(n, arg_v)
        out += arg_v * Bf.d1_hankelh1(n, arg_v)
        out *= -self.beta_n(n)
        # grad(phi)
        out += self.alpha_n(n) * Bf.hankelh1(n, self.cls.k_f * r)

        out *= self.cls.A_in(n) / r
        return out

    # ------------------

    def radial_particle_velocity(self, r, theta, t, mode):
        if mode == 1 or mode is None:
            return self.particle_velocity(t) * cos(theta)
        else:
            return 0

    def tangential_particle_velocity(self, r, theta, t, mode):
        if mode == 1 or mode is None:
            return -self.particle_velocity(t) * sin(theta)
        else:
            return 0

    def particle_velocity(self, t):

        out = Bf.besselj(1, self.x)
        out += self.alpha_n(1) * Bf.hankelh1(1, self.x)
        out -= 2 * self.beta_n(1) * Bf.hankelh1(1, self.x_v)

        out *= self.cls.A_in(1) * self.cls.k_f * self.rho_t
        out /= self.x

        out *= exp(-1j * self.cls.omega * t)

        return out

    # ------------------
    # generic tests
    # ------------------

    def test_properties(self) -> None:
        properties = ["rho_t", "mu_1", "mu_2", "mu_3", "mu_4"]
        self._test_properties(properties)

    def test_methods_n(self) -> None:
        methods = {}
        methods["alpha_n"] = None
        methods["beta_n"] = None
        methods["xi_n"] = None
        methods["gamma_n"] = None
        methods["potential_coefficient"] = None

        self._test_methods_n(methods, threshold=1e-10)

    # ------------------

    def alpha_n(self, n: int) -> complex:
        if n > 1:
            out = n * (n + 1)
            out *= Bf.besselj(n, self.x) * Bf.hankelh1(n, self.x_v)

            out -= self.x * self.gamma_n(n) * Bf.d1_besselj(n, self.x)

            out /= self.xi_n(n)

        elif n > 0:
            out = -2 * (1 - self.rho_t) ** 2 * Bf.besselj(1, self.x)
            out *= Bf.hankelh1(1, self.x_v)

            out -= self.mu_1 * self.mu_3
            out /= self.mu_4

        else:
            out = -Bf.besselj(1, self.x) / Bf.hankelh1(1, self.x)

        return out

    def beta_n(self, n: int) -> complex:
        if n > 1:
            out = Bf.d1_besselj(n, self.x) * Bf.hankelh1(n, self.x)
            out -= Bf.besselj(n, self.x) * Bf.d1_hankelh1(n, self.x)
            out *= -self.x / self.xi_n(n)
        elif n > 0:
            out = self.mu_1 * Bf.hankelh1(1, self.x)
            out -= self.mu_2 * Bf.besselj(1, self.x)
            out *= -(1 - self.rho_t) / self.mu_4
        else:
            out = 0

        return out

    def potential_coefficient(self, n: int) -> complex:
        return self.cls.alpha_n(n) * self.cls.field.A_in(n)

    def xi_n(self, n: int) -> complex:
        out = -n * (n + 1)
        out *= Bf.hankelh1(n, self.x)
        out *= Bf.hankelh1(n, self.x_v)

        out += self.x * Bf.d1_hankelh1(n, self.x) * self.gamma_n(n)

        return out

    def gamma_n(self, n: int) -> complex:
        out = self.x_v * Bf.d1_hankelh1(n, self.x_v)
        out += Bf.hankelh1(n, self.x_v)
        return out

    # ------------------

    @property
    def mu_1(self) -> complex:
        out = -self.x * Bf.d1_besselj(1, self.x)
        out += self.rho_t * Bf.besselj(1, self.x)
        return out

    @property
    def mu_2(self) -> complex:
        out = -self.x * Bf.d1_hankelh1(1, self.x)
        out += self.rho_t * Bf.hankelh1(1, self.x)
        return out

    @property
    def mu_3(self) -> complex:
        out = 1 - 2 * self.rho_t
        out *= Bf.hankelh1(1, self.x_v)

        out += self.x_v * Bf.d1_hankelh1(1, self.x_v)
        return out

    @property
    def mu_4(self) -> complex:
        out = 2 * (1 - self.rho_t) ** 2
        out *= Bf.hankelh1(1, self.x)
        out *= Bf.hankelh1(1, self.x_v)

        out += self.mu_2 * self.mu_3
        return out


class TestScatteringCompareToKing(
    TestBaseDoinikov1994,
    HelperCompareScattering,
):
    def setUp(self) -> None:
        super().setUp()

        self._scattering_compare_threshold = 1e-2
        self.parameters.N_max = 3
        self.parameters.eta_f *= 1e-5
        self.parameters.zeta_f *= 1e-5
        self.parameters.R_0 = 1e-6

        self.cls = SolutionFactory().doinikov_1994_rigid_scattering()
        self.compare_cls = SolutionFactory().king_1934_scattering()


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Sequence

import numpy as np

from osaft.core.functions import (
    cartesian_2_spherical_coordinates as c2s_coord,
)
from osaft.core.functions import (
    spherical_2_cartesian_coordinates as s2c_coord,
)

NDArray = np.ndarray


class BaseGrid(ABC):
    """Base class for coordinate grids

    :param r_max: characteristic length of grid
    :param resolution: number of points, if tuple resolution in each direction
    :param is_upper: if ``True`` only grid in upper half of cartesian plane
    """

    _mesh_x: NDArray
    _mesh_z: NDArray
    _mesh_r: NDArray
    _mesh_theta: NDArray

    def __init__(
        self,
        r_max: float,
        resolution: int | tuple[int, int],
        is_upper: bool,
    ) -> None:

        self._upper = is_upper
        self._r_max = r_max
        self._res_x1, self._res_x2 = self.unpack_resolution(resolution)
        self._compute_mesh()

    # -------------------------------------------------------------------------
    # __init__ helper methods
    # -------------------------------------------------------------------------

    def unpack_resolution(
        self,
        resolution: int | tuple[int, int],
    ) -> tuple[int, int]:
        """Unpacks resolution tuple if needed

        if ``res`` is an `int` a tuple ``(res, res)`` is
        returned. If `res` is a tuple with two values, `res` is
        passed through.
        :param resolution: res of the grid
        """
        if isinstance(resolution, Sequence):
            if len(resolution) == 1:
                if self._upper:
                    return resolution[0], resolution[0] // 2
                else:
                    return resolution[0], resolution[0]
            elif len(resolution) == 2:
                return resolution
            else:
                raise ValueError(
                    "Resolution needs to be either one value for both "
                    "direction or a tuple of two values.",
                )
        else:
            if self._upper:
                return resolution, resolution // 2
            else:
                return resolution, resolution

    # -------------------------------------------------------------------------
    # User API
    # -------------------------------------------------------------------------

    @property
    def mesh_x(self) -> NDArray:
        """Mesh grid array of x-coordinates / ordinate of"""
        return self._mesh_x

    @property
    def mesh_z(self) -> NDArray:
        """Mesh grid array of z-coordinates / abscissa of grid points"""
        return self._mesh_z

    @property
    def mesh_r(self) -> NDArray:
        """Mesh grid array of radial coordinates of grid points"""
        return self._mesh_r

    @property
    def mesh_theta(self) -> NDArray:
        """Mesh grid array of tangential coordinates of grid points"""
        return self._mesh_theta

    @property
    def flat_mesh_x(self) -> NDArray:
        """Flattened mesh array of x-coordinates / ordinate
        grid points"""
        return self.mesh_x.flatten()

    @property
    def flat_mesh_z(self) -> NDArray:
        """Flattened mesh grid array of z-coordinates / abscissa grid points"""
        return self.mesh_z.flatten()

    @property
    def flat_mesh_r(self) -> NDArray:
        """Flattened mesh grid array of radial coordinates of grid points"""
        return self.mesh_r.flatten()

    @property
    def flat_mesh_theta(self) -> NDArray:
        """Flattened mesh grid array of tang. coordinates of grid points"""
        return self.mesh_theta.flatten()

    # -------------------------------------------------------------------------
    # Methods
    # -------------------------------------------------------------------------

    @abstractmethod
    def _compute_mesh(self) -> None:
        pass  # pragma: no coverage


class PolarGrid(BaseGrid):
    """Polar coordinate grid for plotting

    Polar grid for radial coordinate between ``r_min`` and ``r_max``.

    :param r_min: smallest radius of polar grid
    :param r_max: largest radius of polar grid
    :param resolution: number of points, if tuple resolution in each direction
    :param is_upper: if ``True`` only grid in upper half of cartesian plane
    """

    def __init__(
        self,
        r_min: float,
        r_max: float,
        resolution: int | tuple[int, int],
        is_upper: bool,
    ) -> None:

        self._r_min = r_min
        super().__init__(r_max, resolution, is_upper)

    # -------------------------------------------------------------------------
    # API
    # -------------------------------------------------------------------------

    @property
    def arr_r(self) -> NDArray:
        """Linearly spaced array in radial direction"""
        return np.linspace(self._r_min, self._r_max, self.res_r)

    @property
    def arr_theta(self) -> NDArray:
        """Linearly spaced array in tangential direction"""
        if self._upper:
            return np.linspace(0, np.pi, self.res_theta)
        else:
            return np.linspace(-np.pi, np.pi, self.res_theta)

    # -------------------------------------------------------------------------
    # Properties
    # -------------------------------------------------------------------------

    @property
    def res_r(self) -> int:
        return self._res_x1

    @property
    def res_theta(self) -> int:
        return self._res_x2

    @property
    def dr(self) -> float:
        """Step size in radial direction"""
        return self.arr_r[1] - self.arr_r[0]

    @property
    def dtheta(self) -> float:
        """Step size in tangential direction"""
        return self.arr_theta[1] - self.arr_theta[0]

    # -------------------------------------------------------------------------
    # Methods
    # -------------------------------------------------------------------------

    def _compute_mesh(self) -> None:
        self._mesh_theta, self._mesh_r = np.meshgrid(
            self.arr_theta,
            self.arr_r,
        )
        self._mesh_x, self._mesh_z = s2c_coord(self._mesh_r, self._mesh_theta)


class PolarFluidGrid(PolarGrid):
    """Polar coordinate grid for the fluid domain for plotting

    :param R_0: scatterer radius
    :param r_max: largest radius of polar grid
    :param resolution: number of points, if tuple resolution in each direction
    :param is_upper: if ``True`` only grid in upper half of cartesian plane
    """

    def __init__(
        self,
        R_0: float,
        r_max: float,
        resolution: int | tuple[int, int],
        is_upper: bool,
    ) -> None:

        super().__init__(R_0, r_max, resolution, is_upper)


class PolarScattererGrid(PolarGrid):
    """Polar coordinate grid for the fluid domain for plotting

    :param R_0: scatterer radius
    :param resolution: number of points, if tuple resolution in each direction
    :param is_upper: if ``True`` only grid in upper half of cartesian plane
    """

    def __init__(
        self,
        R_0: float,
        resolution: int | tuple[int, int],
        is_upper: bool,
    ) -> None:

        super().__init__(1e-30, R_0, resolution, is_upper)


class CartesianGrid(BaseGrid):
    """Cartesian coordinate grid for plotting

    Cartesian grid for horizontal coordinate between ranging from ``-r_max`` to
    ``r_max`` and vertical coordinate ranging from ``0`` to ``r_max`` or from
    ``-r_max`` to ``r_max`` (see ``upper``).

    :param r_max: characteristic length of grid
    :param resolution: number of points, if tuple resolution in each direction
    :param is_upper: if ``True`` only grid in upper half of cartesian plane
    :param use_offset: if ``True`` center point is slightly offset,
       to avoid singularity
    """

    def __init__(
        self,
        r_max: float,
        resolution: int | tuple[int, int],
        is_upper: bool,
        use_offset: bool,
    ) -> None:
        """Constructor method"""

        self._offset = use_offset
        super().__init__(r_max, resolution, is_upper)

    # -------------------------------------------------------------------------
    # API
    # -------------------------------------------------------------------------

    @property
    def arr_x(self) -> NDArray:
        """Linearly spaced array in x-direction"""
        if self._upper:
            return np.linspace(0, self._r_max, self.res_x)
        else:
            return np.linspace(-self._r_max, self._r_max, self.res_x)

    @property
    def arr_z(self) -> NDArray:
        """Linearly spaced array in z-direction

        if ``offset`` is ``True`` a slight offset is added to the center
        point to avoid possible singularities
        """
        arr = np.linspace(-self._r_max, self._r_max, self.res_z)
        if self.res_z % 2 and self._offset:
            arr[self.res_z // 2] = 1e-30
        return arr

    # -------------------------------------------------------------------------
    # Properties
    # -------------------------------------------------------------------------

    @property
    def res_x(self) -> int:
        return self._res_x2

    @property
    def res_z(self) -> int:
        return self._res_x1

    @property
    def dx(self) -> float:
        return self.arr_x[1] - self.arr_x[0]

    @property
    def dz(self) -> float:
        return self.arr_z[1] - self.arr_z[0]

    # -------------------------------------------------------------------------
    # Methods
    # -------------------------------------------------------------------------

    def _compute_mesh(self) -> None:
        self._mesh_z, self._mesh_x = np.meshgrid(self.arr_z, self.arr_x)
        self._mesh_r, self._mesh_theta = c2s_coord(self.mesh_x, self.mesh_z)


class MaskedCartesianGrid(CartesianGrid):
    """Masked cartesian coordinate grid for plotting

    Cartesian grid for horizontal coordinate ranging from ``-r_max`` to
    ``r_max`` and vertical coordinate ranging from ``0`` to ``r_max`` or from
    ``-r_max`` to ``r_max`` (see ``upper``). Either the inside of a circle
    of radius ``r_mask`` is masked (``inner = False``) or the outside is
    masked (``inner = True``).

    :param r_mask: radius of the masking circle
    :param r_max: characteristic length of grid
    :param resolution: number of points, if tuple resolution in each direction
    :param is_upper: if ``True`` only grid in upper half of cartesian plane
    :param inner: if ``True`` grid points inside circle, else outside
    :param use_offset: if ``True`` center point is slightly offset,
       to avoid singularity
    """

    def __init__(
        self,
        r_mask: float,
        r_max: float,
        resolution: int | tuple[int, int],
        is_upper: bool,
        inner: bool,
        use_offset: bool,
    ) -> None:
        self._inner = inner
        self._r_mask = r_mask
        super().__init__(r_max, resolution, is_upper, use_offset)

    # -------------------------------------------------------------------------
    # Methods
    # -------------------------------------------------------------------------

    def _compute_mesh(self) -> None:
        super()._compute_mesh()
        self._mask_grid()

    def _mask_grid(self) -> None:
        R = np.hypot(self._mesh_z, self._mesh_x)
        if self._inner:
            bools_cartesian = R <= self._r_mask
            bools_polar = self.mesh_r <= self._r_mask
        else:
            bools_cartesian = R >= self._r_mask
            bools_polar = self.mesh_r >= self._r_mask

        self._mesh_z = self._mesh_z[bools_cartesian]
        self._mesh_x = self._mesh_x[bools_cartesian]

        self._mesh_r = self._mesh_r[bools_polar]
        self._mesh_theta = self._mesh_theta[bools_polar]


class CartesianFluidGrid(MaskedCartesianGrid):
    """Masked cartesian coordinate grid of the fluid domain for plotting

    Cartesian grid for horizontal coordinate ranging from ``-r_max`` to
    ``r_max`` and vertical coordinate ranging from ``0`` to ``r_max`` or from
    ``-r_max`` to ``r_max`` (see ``upper``). The points inside the
    scatterer are masked/deleted.

    :param R_0: radius of the masking circle
    :param r_max: dimension of the grid
    :param resolution: number of points, if tuple resolution in each direction
    :param is_upper: if ``True`` only grid in upper half of cartesian plane
    """

    def __init__(
        self,
        R_0: float,
        r_max: float,
        resolution: int | tuple[int, int],
        is_upper: bool,
    ) -> None:

        super().__init__(R_0, r_max, resolution, is_upper, False, False)


class CartesianScattererGrid(MaskedCartesianGrid):
    """Masked cartesian coordinate grid of the scatterer domain for plotting

    Cartesian grid for horizontal coordinate ranging from ``-r_max`` to
    ``r_max`` and vertical coordinate ranging from ``0`` to ``r_max`` or from
    ``-r_max`` to ``r_max`` (see ``upper``). The points outside the
    scatterer are masked/deleted.

    :param R_0: radius of the scatterer
    :param resolution: number of points, if tuple resolution in each direction
    :param is_upper: if ``True`` only grid in upper half of cartesian plane
    :param use_offset: if ``True`` center point is slightly offset,
       to avoid singularity
    """

    def __init__(
        self,
        R_0: float,
        resolution: int | tuple[int, int],
        is_upper: bool,
        use_offset: bool = True,
    ) -> None:
        super().__init__(R_0, R_0, resolution, is_upper, True, use_offset)


if __name__ == "__main__":
    pass



# Import logger
from osaft.core.logger import log

# Import API
from osaft.core.functions import pi
from osaft.core.backgroundfields import BackgroundField, WaveType
from osaft.core.fluids import InviscidFluid, ViscoelasticFluid, ViscousFluid
from osaft.core.solids import ElasticSolid, RigidSolid
from osaft.plotting.arf.arf_plots import ARFPlot
from osaft.plotting.scattering.fluid_plots import FluidScatteringPlot
from osaft.plotting.scattering.particle_plots import (
    ParticleScatteringPlot,
    ParticleWireframePlot,
)

# IMPORT API solutions
import osaft.solutions.gorkov1962 as gorkov1962
import osaft.solutions.king1934 as king1934
import osaft.solutions.yosioka1955 as yosioka1955
import osaft.solutions.doinikov1994rigid as doinikov1994rigid
import osaft.solutions.doinikov1994compressible as doinikov1994compressible
import osaft.solutions.settnes2012 as settnes2012
import osaft.solutions.hasegawa1969 as hasegawa1969
import osaft.solutions.doinikov2021viscous as doinikov2021viscous
import osaft.solutions.doinikov2021viscoelastic as doinikov2021viscoelastic

# Legacy names of modules
Gorkov1962 = gorkov1962
King1934 = king1934
Yosioka1955 = yosioka1955
Doinikov1994Rigid = doinikov1994rigid
Doinikov1994Compressible = doinikov1994compressible
Settnes2012 = settnes2012
Doinikov2021Viscous = doinikov2021viscous
Doinikov2021Viscoelastic = doinikov2021viscoelastic
Hasegawa1969 = hasegawa1969



from __future__ import annotations

from abc import ABC

import numpy as np

from osaft import BackgroundField, ViscousFluid
from osaft.core.functions import conj, pi, sin
from osaft.core.geometries import Sphere
from osaft.core.variable import ActiveVariable, PassiveVariable
from osaft.solutions.base_arf import BaseARF


class BaseARFLimiting(BaseARF, ABC):
    """Abstract base class for the computation of the acoustic radiation
    in the models by Doinikov in limiting cases.
    """

    # Required attributes of child class
    _small_boundary_layer: PassiveVariable
    _large_boundary_layer: PassiveVariable
    _background_streaming: PassiveVariable
    field: BackgroundField
    fluid: ViscousFluid
    sphere: Sphere

    # Required properties of child class
    abs_pos: float
    delta: float
    k_f: complex
    R_0: float
    rho_f: float
    x: complex
    x_v: complex

    def __init__(self):

        # Dependent variables
        self._norm_delta = ActiveVariable(
            self._compute_norm_delta,
            "normalized boundary layer thickness",
        )
        self._norm_delta.is_computed_by(
            self.fluid._delta,
            self.sphere._R_0,
        )

    # -------------------------------------------------------------------------
    # Getters and Setters for independent variables
    # -------------------------------------------------------------------------

    @property
    def small_boundary_layer(self) -> bool:
        """Use limiting case of a small viscous boundary layer :math:`\\delta`

        :getter: returns if a small viscous boundary layer case is used
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._small_boundary_layer.value

    @small_boundary_layer.setter
    def small_boundary_layer(self, value):
        self._small_boundary_layer.value = value

    @property
    def large_boundary_layer(self) -> bool:
        """Use limiting case of a large viscous boundary layer :math:`\\delta`

        :getter: returns if a large viscous boundary layer case is used
        :setter: automatically invokes
            :func:`osaft.core.variable.BaseVariable.notify`
        """
        return self._large_boundary_layer.value

    @large_boundary_layer.setter
    def large_boundary_layer(self, value):
        self._large_boundary_layer.value = value

    @property
    def background_streaming(self) -> bool:
        """Background streaming contribution to the ARF

        :getter: returns if background streaming is considered
        :setter: automatically invokes
            :func:`osaft.core.variable.BaseVariable.notify`
        """
        return self._background_streaming.value

    @background_streaming.setter
    def background_streaming(self, value):
        self._background_streaming.value = value

    # -------------------------------------------------------------------------
    # Background Streaming Contribution
    # -------------------------------------------------------------------------

    def _F_2_standing(self) -> float:
        # F2 according to (5.17)
        x_con = self.x.conjugate()

        F2 = self.x - x_con
        F2 *= sin(self.abs_pos * (self.k_f + conj(self.k_f)))
        F2 *= sin(self.x + x_con) / (self.x + x_con)

        tmp = self.x + x_con
        tmp *= sin(self.abs_pos * (self.k_f - conj(self.k_f)))

        # prevent division by zero error
        if self.x.imag > 1e-5:  # pragma: no cover
            tmp *= np.sinh(2 * self.x.imag) / (2 * self.x.imag)

        F2 -= tmp
        F2 *= -self.x * x_con / self.x_v**2
        F2 *= self.rho_f * pi * 0.75 * self.field.abs_A_squared

        return F2.real

    def _F_2_travelling(self) -> float:
        # F2 according to (5.12)
        x_con = self.x.conjugate()
        F2 = -self.x * x_con * (self.x + x_con) * 1j / self.x_v**2
        F2 *= sin(self.x - x_con) / (self.x - x_con)
        F2 *= self.rho_f * pi * 1.5 * self.field.abs_A_squared
        return F2.real


if __name__ == "__main__":
    pass
